#========================================================================================
# package      DE_EPAGES::Order::API::Object::ContainerPaymentMethod
# state        public
# base         DE_EPAGES::Payment::API::Object::PaymentMethod
# description  object interface for PaymentMethod to handle features hire purchase and payment discounts.
#========================================================================================
package DE_EPAGES::Order::API::Object::ContainerPaymentMethod;
use base qw ( DE_EPAGES::Payment::API::Object::PaymentMethod );

use strict;
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( LogError );
use DE_EPAGES::Customer::API::Shop qw ( GetCustomerForUser );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias );
use DE_EPAGES::Order::API::Object::LineItemPaymentDiscount;
use DE_EPAGES::Order::API::PaymentMethodFilter;

use constant VARCHAR_255 => 255;

#========================================================================================
# function     lineItem
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $hInfo = $PaymentMethod->lineItem($Container);
#----------------------------------------------------------------------------------------
# description  Collect information for line item.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# return       $hInfo | payment line item info, keys
#               <ul>
#                   <li>BasePrice - shipping sub total of container - float
#                   <li>LineItemPrice - additional price for hire purchase - float
#                   <li>Name -  name of payment method
#                   <li>PaymentMethod - self - object
#                   <li>Quantity - quantity is 1 - float
#                   <li>TaxClass - tax class of payment - object
#                   <li>TaxRate - tax rate - float
#               </ul> | ref.hash
#========================================================================================
sub lineItem {
    my $self = shift;
    my ($Container) = @_;

    my $ExistingTaxClass = $self->get('TaxClass');
    my $TaxClass = $ExistingTaxClass ? $Container->get('TaxMatrix')->getTaxClassByAlias($ExistingTaxClass->alias) : undef;
    my $TaxRate = defined $TaxClass ? $Container->taxRate($TaxClass) : 0;

    # EPG-25798: Can't add a product to the basket if the name of the default payment method
    #            is longer than 255 characters
    my $PaymentMethodName = $self->get('Name', $Container->get('LanguageID'));
    my $Postfix = '...';
    $PaymentMethodName = substr ($PaymentMethodName, 0, VARCHAR_255 - length $Postfix) . $Postfix
        if (defined $PaymentMethodName && length $PaymentMethodName > VARCHAR_255);

    my $hInfo = $Container->get([qw (CurrencyID TaxModel TaxMatrix)]);

    return {
        'Name'              => $PaymentMethodName,
        'PaymentMethod'     => $self,
        'Quantity'          => 1,
        'TaxClass'          => $TaxClass,
        'TaxRate'           => $TaxRate,
        'LineItemPrice'     => $self->price({
            'CurrencyID' => $hInfo->{'CurrencyID'},
            'TaxModel'   => $hInfo->{'TaxModel'},
            'TaxMatrix'  => $hInfo->{'TaxMatrix'},
            'Container'  => $Container,
        }),

    };
}

sub price {
    my $self = shift;
    my ($hValues) = @_;

    return $self->_price($hValues->{'CurrencyID'}, $hValues->{'TaxModel'});
}

#========================================================================================
# function     _price
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $PaymentPrice = $PaymentMethod->_price($CurrencyID, $TaxModel);
#----------------------------------------------------------------------------------------
# description  Returns the payment price.
#----------------------------------------------------------------------------------------
# input        $CurrencyID | currency id | string(3)
# input        $TaxModel | tax model (0=net, 1=gross) | int
# return       $PaymentPrice | price of payment | money
#========================================================================================
sub _price {
    my $self = shift;
    my ($CurrencyID, $TaxModel) = @_;

    return 0 unless defined $TaxModel;
    return 0 unless defined $CurrencyID;

    my $ahLevels = $self->getLevels($CurrencyID, $TaxModel);
    return 0 unless scalar @$ahLevels;
    return $ahLevels->[0]{'BaseValue'};
}

#========================================================================================
# function     calculatePaymentPrice
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $PaymentPrice = $PaymentMethod->calculatePaymentPrice($Container);
#----------------------------------------------------------------------------------------
# description  Returns the payment price by the given requirements.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# return       $PaymentPrice | price of shipping | money
#========================================================================================
sub calculatePaymentPrice {
    my $self = shift;
    my ($Container) = @_;

    return $self->_price($Container->get('CurrencyID'), $Container->get('TaxModel'));
}

#========================================================================================
# function     canAddToBasket
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PaymentMethod->canAddToBasket($hVars);
#----------------------------------------------------------------------------------------
# description  Tests if a payment method can be added to the basket with the given currency.
#               Uses function <function DE_EPAGES::Payment::API::Object::PaymentMethod::price>
#               to evaluate if payment methode has a price (price can be 0). Possible for
#               methods, which costs, but have no price in the given currency.
#----------------------------------------------------------------------------------------
# input        $hVars | vars of container and basket | ref.hash
# return       $Can | can added to the basket | boolean
#========================================================================================
sub canAddToBasket {
    my $self = shift;
    my ($hVars) = @_;
    return 0 unless $self->get('IsActivated');
    return 0 unless $self->get('PaymentType')->get('IsActivated');
    return 0 unless $self->_canDoCheckoutType();

    my $LineItemContainer = $hVars->{'LineItemContainer'};

    # test regions
    my $Region= $self->get('Region');
    if (defined $Region) {
        $hVars->{'BillingCountryID'} = $LineItemContainer->get('BillingCountryID') if defined $LineItemContainer and not exists $hVars->{'BillingCountryID'};
        if (defined $hVars->{'BillingCountryID'}) {
            return 0 unless grep {$_->{'CountryID'} == $hVars->{'BillingCountryID'}} @{$Region->get('Countries')};
        }
    }

    # test customer groups
    my $CustomerGroups = $self->get('CustomerGroups');
    if (scalar @$CustomerGroups) {
        my $CustomerGroup = $hVars->{'CustomerGroup'};
        if(not defined $CustomerGroup and defined $LineItemContainer) {
            my $ContainerParent = $LineItemContainer->businessContainer->parent;
            my $Customer = $ContainerParent->get('Customer');
            if (not defined $Customer) {
                my $User = $ContainerParent->get('User');
                $Customer = GetCustomerForUser($User, $ContainerParent->getSite()) if defined $User;
            }
            $CustomerGroup = $Customer->get('CustomerGroup') if defined $Customer;
        }
        $CustomerGroup = $CustomerGroups->[0]->getSite()->get('NonMemberCustomerGroup') unless defined $CustomerGroup;
        $hVars->{'CustomerGroup'} = $CustomerGroup;
        return 0 if not defined $CustomerGroup or not grep { $_->id == $CustomerGroup->id } @$CustomerGroups;
    }

    # test if the payment method has a price limit
    if ($self->get('HasPriceLimit') and defined $LineItemContainer) {

        #get the DiscountsSubTotal of lineitemcontainer to check if the
        #DiscountsSubTotal is in the range of pricelimit MinValue and MaxValue

        my $DiscountsSubTotal = $LineItemContainer->get('DiscountsSubTotal');

        my $ContainerCurrencyID = $LineItemContainer->get('CurrencyID');
        my ($hPriceLimit) = grep { $_->{'CurrencyID'} eq $ContainerCurrencyID } @{$self->get('PriceLimits')} ;

        my $MinPriceValue = $hPriceLimit->{'MinPriceValue'};
        my $MaxPriceValue = $hPriceLimit->{'MaxPriceValue'};

        return 0 if (defined $MinPriceValue and ($DiscountsSubTotal < $MinPriceValue));
        return 0 if (defined $MaxPriceValue and ($MaxPriceValue < $DiscountsSubTotal));

    }

    if ($hVars->{'skipUsableAtStorefrontCheck'}) {
        return 1;
    }

    return $self->usableAtStorefront($hVars);
}
#----------------------------------------------------------------------------------------
sub _canDoCheckoutType {
    my $self = shift;

    my $CanDoCheckoutType;
    eval {
        my $PaymentMethodFilter = DE_EPAGES::Order::API::PaymentMethodFilter->new();
        $CanDoCheckoutType = $PaymentMethodFilter->canDoCheckoutType($self);
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogError($Error);
        $CanDoCheckoutType = 1; #If something went wrong, ignore the canDoCheckoutType check
    }
    return $CanDoCheckoutType;
}

#========================================================================================
# function     addToContainer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LineItem = $PaymentMethod->addToContainer($Container);
#----------------------------------------------------------------------------------------
# description  Adds payment and tax line item to container.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# return       $LineItem | line item | object
#========================================================================================
sub addToContainer {
    my $self = shift;
    my ($Container) = @_;

    return $Container->insertLineItem($self->get('PaymentType')->get('LineItemClass'), $self->lineItem($Container));
}

#========================================================================================
# function     recalculateLineItem
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PaymentMethod->recalculateLineItem($LineItem);
#----------------------------------------------------------------------------------------
# description  Called to recalc the line item. And adds a payment discount if possible.
#----------------------------------------------------------------------------------------
# input        $LineItem | line item | object
#========================================================================================
sub recalculateLineItem {
    my $self = shift;
    my ($LineItem) = @_;

    my $Container = $LineItem->container;
    $LineItem->set($self->lineItem($Container));

    #add payment discount
    if ( not defined $Container->get('PaymentDiscount') ) {
        my $hInfo = DE_EPAGES::Order::API::Object::LineItemPaymentDiscount->lineItemInfo($Container);
        $Container->insertLineItem(LoadClassByAlias('LineItemPaymentDiscount'), $hInfo ) if defined $hInfo;;
    }
    return;
}

1;
