#==============================================================================
# §package      DE_EPAGES::Order::UI::JSON
# §state        public
# §base         base DE_EPAGES::Presentation::UI::JSON
# §description  JSON actions to e.g. generate order documents, save customizable
#               basket buttons.
#==============================================================================
package DE_EPAGES::Order::UI::JSON;
use base DE_EPAGES::Presentation::UI::JSON;

use strict;

use DE_EPAGES::Core::API::File qw ( RelativePath GetSimilarFileName ExistsFile );
use DE_EPAGES::Core::API::PerlTools qw ( isValid GetCurrentDateTime );
use DE_EPAGES::Design::API::Constants qw ( CUSTOMIZABLE_BUTTONS );
use DE_EPAGES::Design::API::Style qw ( ValidateAndCompareToDefaultButtonValues InvalidateAllShopStyles );
use DE_EPAGES::Order::API::OrderDocument qw ( CreateOrderDocument CreateShippingLabel );
use DE_EPAGES::Order::API::PickupTokenExpiration;
use DE_EPAGES::Order::API::PickupTokenValidation;
use DE_EPAGES::Order::UI::Basket_Ext;

use Image::Size qw(imgsize);
use JSON;

use constant MAX_IMAGE_WIDTH => 280;

#========================================================================================
# §function     JSONGenerateDocuments
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $Handler->JSONGenerateDocuments( $Servlet );
#----------------------------------------------------------------------------------------
# §description  JSON ChangeAction which batch-generates documents like packing slips and
#               invoices for a given list of orders.
#----------------------------------------------------------------------------------------
# §input        $Servlet | servlet | object
# §return       $Output | json response | object
#========================================================================================
sub JSONGenerateDocuments {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object;
    my $Form = $Servlet->form;

    my @Orders = $Form->value('OrderList');
    # TODO: Error Feedback
    return if not @Orders or scalar(@Orders) == 0;

    $Form->setlocale($Form->value('Locale'));
    my $DeliveryDate =  $Form->field($Shop, $Form->value('DeliveryDate'),
                                    { 'Type'=>'reg_date', 'Name'=>'DeliveryDate' });
    my $IssueDate =     $Form->field($Shop, $Form->value('IssueDate'), { 'Type'=>'reg_date', 'Name'=>'IssueDate' });

    my $CreateInvoices =        $Form->value('createInvoice') eq 'true' ? 1 : 0;
    my $CreatePackingSlips =    $Form->value('createPackingSlip') eq 'true' ? 1 : 0;
    my $CreateShippingLabels =  $Form->value('createShippingLabel') eq 'true' ? 1 : 0;
    # flags
    my $SetReadyForShippingOn = $Form->value('ReadyForShippingOn') eq 'true' ? 1 : 0;
    my $SetDispatchedOn =       $Form->value('DispatchedOn') eq 'true' ? 1 : 0;
    my $SetInvoicedOn =         $Form->value('InvoicedOn') eq 'true' ? 1 : 0;

    my $NumTotalDocuments = 0;

    foreach my $OrderID ( sort @Orders ){
        my $Order = $Shop->loadObject($OrderID, 'CustomerOrder');

        # create invoice, but skip if former invoice already exists
        if(  $CreateInvoices and (not $Order->findChild('Invoices')
          or scalar(@{$Order->child('Invoices')->children}) == 0)) {
            CreateOrderDocument(
                $Order,
                'Invoice',
                'Invoices',
                {
                    'DeliveryDate' => $DeliveryDate,
                    'IssueDate' => $IssueDate,
                    'UseAlwaysDeliveryDate' => 1
                }
            );
        }

        # create packing slip, but skip if former packing slip exists
        if(  $CreatePackingSlips and (not $Order->findChild('PackingSlips')
          or scalar(@{$Order->child('PackingSlips')->children}) == 0)) {
            CreateOrderDocument(
                $Order,
                'PackingSlip',
                'PackingSlips',
                {
                    'DeliveryDate' => $DeliveryDate,
                    'IssueDate' => $IssueDate,
                    'UseAlwaysDeliveryDate' => 1
                }
            );
        }

        # create shipping label, but skip if former shipping label exists
        if( $CreateShippingLabels
            and (   not $Order->findChild('ShippingLabels')
                    or scalar(@{$Order->child('ShippingLabels')->children}) == 0))
            {
                my $FulfillmentMethod = $Order->get('ShippingFulfillmentMethod');

                if (defined $FulfillmentMethod) {
                    my $AutoGenerateLabels = $FulfillmentMethod->get('AutoGenerateLabels');

                    # We need special check if document creation is triggered from the batch action menu
                    # do not create shipping labels when triggered from batchaction and label creation setting is "no"
                    if (!$Shop->isUnity() && $AutoGenerateLabels == 0) {
                        # Do not create Shipping labels for Base shops when AutoGenerateLabels is set to "no"
                    } else {
                        CreateShippingLabel($Order);
                    }
                }
            }

        # update order states?
        # this implicitly sends mail to the customer if activated
        $Order->set({'ReadyForShippingOn' => GetCurrentDateTime()}) if $SetReadyForShippingOn;
        $Order->set({'DispatchedOn' => GetCurrentDateTime()}) if $SetDispatchedOn;
        $Order->set({'InvoicedOn' => GetCurrentDateTime()}) if $SetInvoicedOn;

        $NumTotalDocuments += scalar(@{$Order->get('Documents')});
    }

    # return the new number of documents of the order
    return $self->writeContentOutput({
        'OrderDocumentCount'=> $NumTotalDocuments
    });
}

#========================================================================================
# §function     JSONSaveCustomizableBasketButton
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       JSONSaveCustomizableBasketButton($Servlet);
#----------------------------------------------------------------------------------------
# §description  Method to save customizable buttons settings.
#               Triggers hook to save settings for the customizable basket button.
#----------------------------------------------------------------------------------------
# §input        $Servlet | servlet object | object
# §return       $Output | json response | object
#========================================================================================
sub JSONSaveCustomizableBasketButton {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object;
    my $Form = $Servlet->form;

    my $DefaultStyle = $Shop->get('DefaultStyle');

    # Get allowed button styling attributes.
    my $hCUSTOMIZABLEBUTTONS = CUSTOMIZABLE_BUTTONS();
    # Get settings.
    my $hCustomizableBasketButton = from_json($Form->value('CustomizableBasketButton'));

    # Validate values.
    my $hDefaults = $DefaultStyle->get('DefaultBasketButtonStyle');
    $hDefaults = $DefaultStyle->get('DefaultButtonStyle') unless defined $hDefaults;

    my $hCustomizableBasketButtonForJSON->{'CustomizableBasketButton'} =
        ValidateAndCompareToDefaultButtonValues(
            $Form,
            $hCustomizableBasketButton,
            $hCUSTOMIZABLEBUTTONS,
            $hDefaults
        );

    if ($Form->existsFormError()) {
        my $aFormErrors = $Form->formErrors();
        foreach my $hFormError (@$aFormErrors) {
            $hFormError->{'Form'} = 'CustomizableBasketButton';
        }
    }
    $Form->executeFormError();

    # Store image settings.
    my $UploadedImagePath = $hCustomizableBasketButton->{'Image'}->{'Path'};
    my $AbsoluteUploadedImagePath = $ENV{'EPAGES_SHARED'}.$UploadedImagePath;

    if(ExistsFile($AbsoluteUploadedImagePath)) {
        # Get image resolution.
        my ($ImageWidth, $ImageHeight) = imgsize($AbsoluteUploadedImagePath);
        # Check if image should be resized.
        if($ImageWidth > MAX_IMAGE_WIDTH) {
            my $MediaGallery = $Shop->get('MediaGallery');
            my $RelativeImagePath = RelativePath($MediaGallery->get('PublicPath'), $AbsoluteUploadedImagePath);

            # Generate a new image name for resized image version.
            $UploadedImagePath = GetSimilarFileName($AbsoluteUploadedImagePath);
            $MediaGallery->resizeImage($RelativeImagePath, RelativePath($MediaGallery->get('PublicPath'), $UploadedImagePath), MAX_IMAGE_WIDTH);

            # Get resolution of the new image.
            ($ImageWidth, $ImageHeight) = imgsize($UploadedImagePath);

            # Set path relative to "Shared" folder.
            $UploadedImagePath = '/'.RelativePath($ENV{'EPAGES_SHARED'}, $UploadedImagePath);
        }
        $hCustomizableBasketButtonForJSON->{'CustomizableBasketButton'}->{'Image'} = {
            'Path' =>  $UploadedImagePath,
            'Height' => $ImageHeight,
            'Width' => $ImageWidth,
        };
        $hCustomizableBasketButtonForJSON->{'CustomizableBasketButton'}->{'IsCustomized'} = 1;
    } else {
        $hCustomizableBasketButtonForJSON->{'CustomizableBasketButton'}->{'Image'} = {
            'Path' =>  undef,
            'Height' =>undef,
            'Width' => undef,
        };
    }

    $DefaultStyle->set($hCustomizableBasketButtonForJSON);
    my $hParams = {
        'ResponseContent' => $hCustomizableBasketButtonForJSON,
        'Style' => $DefaultStyle,
        'Form' => $Form,
    };
    InvalidateAllShopStyles($Shop);
    return $self->writeContentOutput($hParams->{'ResponseContent'});
}

#========================================================================================
# §function     JSONGetSessionForBasket
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       JSONGetSessionForBasket($Servlet);
#----------------------------------------------------------------------------------------
# §description  Method for generating a user session based on the basket pickup token
#               The session user is then the basket user
#----------------------------------------------------------------------------------------
# §input        $Servlet | servlet object | object
# §return       $Output | json response | object
#========================================================================================
sub JSONRegisterSession {
    my $self = shift;
    my ($Servlet) = @_;


    my $hParams = {};
    my $PickupToken = $Servlet->requestHeaders('X-PickupToken');
    if(! isValid($PickupToken)) {
        $PickupToken = $Servlet->params('PickupToken');
    }
    # verify pickup token here
    my $Basket            = $Servlet->object();
    my $OrderCreationDate = $Basket->get('CreationDate');
    my $PickupTokenExpiration = $self->_createPickupTokenExpiration();
    my $IsExpired             = $PickupTokenExpiration->isExpired($OrderCreationDate);
    my $PickupTokenValidation = $self->_createPickupTokenValidation();
    my $UrlForValidation      = $PickupTokenValidation->getUrlForPickupTokenValidation($Servlet);
    my $IsTokenValid          = $PickupTokenValidation->checkPickupToken($PickupToken, $UrlForValidation);

    if ($IsTokenValid && !$IsExpired) {
        # register a new session if needed
         ##no critic (ProtectPrivateSubs)
        $hParams = DE_EPAGES::Order::UI::Basket_Ext::_RegisterSession($Servlet, $Basket);
    } else {
        $hParams->{'TokenInvalidOrExpired'} = 1;
    }

    return $self->writeContentOutput($hParams);
}

sub _createPickupTokenExpiration {
    return DE_EPAGES::Order::API::PickupTokenExpiration->new();
}

sub _createPickupTokenValidation {
    return DE_EPAGES::Order::API::PickupTokenValidation->new();
}



1;