#========================================================================================
# package      DE_EPAGES::PaymentTypes::API::Object::PaymentMethodFixPrice
# state        public
# base         DE_EPAGES::Order::API::Object::ContainerPaymentMethod
# description  Object interface for fix price payment.
#========================================================================================
package DE_EPAGES::PaymentTypes::API::Object::PaymentMethodFixPrice;
use base DE_EPAGES::Order::API::Object::ContainerPaymentMethod;

use strict;
use DE_EPAGES::Core::API::Error qw (Error );

sub ErrorDefinition { return {
    'PaymentMethodNoPrice'    => 'payment method ( #PaymentMethodID ) has no price for currency #CurrencyID.',
}}

#========================================================================================
# function     price
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PaymentPrice = $Object->price( $hVars );
#----------------------------------------------------------------------------------------
# description  Returns the payment price by the given requirements.
#----------------------------------------------------------------------------------------
# input        $hVars | <ul>
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price    - int
#               </ul> | ref.hash
# return       $PaymentPrice | price of payment | money
#========================================================================================
sub price {
    my $self = shift;
    my ($hValues) = @_;

    return $self->_price($hValues->{'CurrencyID'}, $hValues->{'TaxModel'},$hValues->{'TaxMatrix'});
}

#========================================================================================
# function     calculatePaymentPrice
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $PaymentPrice = $PaymentMethod->calculatePaymentPrice($Container);
#----------------------------------------------------------------------------------------
# description  Returns the payment price by the given requirements.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# return       $PaymentPrice | price of shipping | money
#========================================================================================
sub calculatePaymentPrice {
    my $self = shift;
    my ($Container) = @_;

    my $hInfo = $Container->get([qw (CurrencyID TaxModel TaxMatrix)]);
    return $self->_price($hInfo->{'CurrencyID'}, $hInfo->{'TaxModel'}, $hInfo->{'TaxMatrix'});
}

#========================================================================================
# function     _price
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $PaymentPrice = $PaymentMethod->_price($CurrencyID, $TaxModel);
#----------------------------------------------------------------------------------------
# description  Returns the shipping price.
#----------------------------------------------------------------------------------------
# input        $CurrencyID | currency id | string(3)
# input        $TaxModel | tax model (0=net, 1=gross) | int
# return       $PaymentPrice | price of payment | money
#========================================================================================
sub _price {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $TaxMatrix) = @_;

    my $ahLevels = $self->getLevels($CurrencyID, $TaxModel, {'TaxMatrix' => $TaxMatrix});
    Error('PaymentMethodNoPrice', {
        'PaymentMethodID' => $self->id,
        'CurrencyID' => $CurrencyID,
    }) unless scalar @$ahLevels;
    return $ahLevels->[0]{'BaseValue'};
}

1;
