#========================================================================================
# package      DE_EPAGES::Presentation::t::CheckCaptcha
# state        public
#----------------------------------------------------------------------------------------
# description  Provides tests for Google Captcha.
#========================================================================================
package DE_EPAGES::Presentation::t::CheckCaptcha;
use base DE_EPAGES::Core::API::Test;

use strict;
use DE_EPAGES::Presentation::UI::Object;

use Test::More;
use Test::Exception;;
use Test::Mockify;
use Test::Mockify::Verify;
use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use DE_EPAGES::Test::Mock::MockifyServlet;
use Test::MockObject::Extends;
use DE_EPAGES::Test::Mock::EpagesObject;

use JSON;

sub test {
    Test_CheckWrongCaptcha();
    Test_CheckTrueCaptcha();
    Test_CheckFalseSuccessCodeCaptcha();
    Test_CheckFalseHostCaptcha();
    Test_ValidResponseCodeCaptcha();
    Test_NotValidResponseCodeCaptcha();
    Test_NoPrivateKeyCaptcha();
    Test_PrivateKeyCaptcha();
    done_testing();
}

sub Test_CheckWrongCaptcha {
    my $SubTestName = (caller(0))[3];
    my $Servlet = _createServlet();
    my $GoogleResponse = {};

    dies_ok(
        sub {
            DE_EPAGES::Presentation::UI::Object->CheckGoogleCaptcha($Servlet, $GoogleResponse);
        }, "$SubTestName - False GoogleCaptcha"
    );
}

sub Test_CheckTrueCaptcha {
    my $SubTestName = (caller(0))[3];
    my $Servlet = _createServlet();
    my $GoogleResponse = {
        'hostname' => 'test.vm-intern.epages.com',
        'success' => JSON::true,
    };

    lives_ok(
        sub {
            DE_EPAGES::Presentation::UI::Object->CheckGoogleCaptcha($Servlet, $GoogleResponse)
        }, "$SubTestName - True GoogleCaptcha"
    )
}

sub Test_CheckFalseHostCaptcha {
    my $SubTestName = (caller(0))[3];
    my $Servlet = _createServlet();
    my $GoogleResponse = {
        'hostname' => 'false.vm-intern.epages.com',
    };

    dies_ok(
        sub {
            DE_EPAGES::Presentation::UI::Object->CheckGoogleCaptcha($Servlet, $GoogleResponse)
        }, "$SubTestName - False hostname GoogleCaptcha"
    )
}

sub Test_CheckFalseSuccessCodeCaptcha {
    my $SubTestName = (caller(0))[3];
    my $Servlet = _createServlet();
    my $GoogleResponse = {
        'hostname' => 'test.vm-intern.epages.com',
        'success' => JSON::false,
    };

    dies_ok(
        sub {
            DE_EPAGES::Presentation::UI::Object->CheckGoogleCaptcha($Servlet, $GoogleResponse)
        }, "$SubTestName - False success code GoogleCaptcha"
    )
}

sub Test_ValidResponseCodeCaptcha {
    my $SubTestName = (caller(0))[3];

    my $ValidResponseCode = "valid-response-code";
    my $Servlet = _createServlet($ValidResponseCode);

    lives_ok(
        sub {
            my $MockedPresentationObject = _createMockedPresentationObject($Servlet);
            $MockedPresentationObject->GoogleCaptchaCall($Servlet)
        }, "$SubTestName - Valid Response code GoogleCaptcha"
    )
}

sub Test_NotValidResponseCodeCaptcha {
    my $SubTestName = (caller(0))[3];
    my $Servlet = _createServlet();

    dies_ok(
        sub {
            my $MockedPresentationObject = _createMockedPresentationObject($Servlet);
            $MockedPresentationObject->GoogleCaptchaCall($Servlet)
        }, "$SubTestName - Not Valid Response code GoogleCaptcha"
    )
}

sub Test_NoPrivateKeyCaptcha {
    my $SubTestName = (caller(0))[3];
    my $Servlet = _createServlet();

    lives_ok(
        sub {
            my $MockedPresentationObject = _createMockedPresentationObject();
            $MockedPresentationObject->GoogleCaptcha($Servlet)
        }, "$SubTestName - No Private Key GoogleCaptcha"
    )
}

sub Test_PrivateKeyCaptcha {
    my $SubTestName = (caller(0))[3];
    my $Servlet = _createServlet();

    my $PrivateKey = "valid-private-key";

    dies_ok(
        sub {
            my $MockedPresentationObject = _createMockedPresentationObject($PrivateKey);
            $MockedPresentationObject->GoogleCaptcha($Servlet)
        }, "$SubTestName - Private Key GoogleCaptcha"
    )
}

#----------------------------------------------------------------------------------------
sub _createForm {

    my $MockedForm = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::Presentation::API::Form');

    $MockedForm->mock('executeFormError', sub {
        my $self = shift;
        my ($hErrorInfo) = @_;

        my $ErrorReason = $hErrorInfo->{'Reason'};

        if ($ErrorReason eq "RECAPTCHA_WRONG") {
            die;
        }
    });

    return $MockedForm->getMockObject();
}

#----------------------------------------------------------------------------------------
sub _createServlet {
    my $ResponseCode = shift;
    my $Shop = _createShop();

    my $MockifyServlet = DE_EPAGES::Test::Mock::MockifyServlet->new($Shop);

    $MockifyServlet->addMockWithReturnValue('form', _createForm());
    $MockifyServlet->addMockWithReturnValue('serverVariables', {'HTTP_HOST' => 'test.vm-intern.epages.com'});
    $MockifyServlet->addParams({'g-recaptcha-response' => $ResponseCode});

    return $MockifyServlet->getMockObject();
}

#----------------------------------------------------------------------------------------
sub _createMockedPresentationObject {
    my ( $PrivateKey ) = @_;

    my $PresentationObject = DE_EPAGES::Presentation::UI::Object->new();
    my $PresentationObjectMock = Test::MockObject::Extends->new($PresentationObject);
    $PresentationObjectMock->mock('_loadRootObject', sub {return _createSystemObject($PrivateKey);}); #overwrite LoadRootObject

    return $PresentationObjectMock;
}

#----------------------------------------------------------------------------------------
sub _createShop {

    my $hAttributes = {
        'Alias' =>'DemoShop',
        'IsUnityRegistered' => 0,
        'EnableReCaptcha' => 1,
    };

    my $MockedShop = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Shop::API::Object::Shop', $hAttributes);

    my $Shop  = $MockedShop->getMockObject();
    $Shop->set({'Site' => $Shop});

    return $Shop;
}

sub serverVariables {
    return {
        HTTP_HOST => 'test.vm-intern.epages.com',
    };
}

#========================================================================================
# DAL Wrapper
#========================================================================================
sub _createSystemObject {
    my $PrivateKey = shift;

    my $hSystemAttributes = {
        'EpagesVersion'         => '7.24.0',
        'ProviderID'            => 'epages',
        'StoreName'             => 'Store',
        'ReCaptchaPrivateKey'   => $PrivateKey,
    };
    return DE_EPAGES::Test::Mock::EpagesObject->new('DE_EPAGES::Object::API::Object::Site', $hSystemAttributes);
}

#----------------------------------------------------------------------------------------
__PACKAGE__->runScript();

1;
