#========================================================================================
# package      DE_EPAGES::PriceList::Hooks::ProductService
# state        public
# description  extends PoductService with elements
#========================================================================================
package DE_EPAGES::PriceList::Hooks::ProductService;

use strict;
use DE_EPAGES::PriceList::API::Constants qw( UNITY_PRICELIST_PREFIX );
use DE_EPAGES::Object::API::Factory qw ( LoadObject );
use DE_EPAGES::Core::API::PerlTools qw ( ElementOf );


#========================================================================================
# function     OnProductServiceGetProductData
# state        public
#----------------------------------------------------------------------------------------
# syntax       OnProductServiceGetProductData( $hParams );
#----------------------------------------------------------------------------------------
# description  Hook function for the API_ProductService_GetProductData hook.
#               Extends the ProductService result (hResult) with bulk price informations
#               for the given product.
#----------------------------------------------------------------------------------------
# input        $hParams |<ul>
#                   <li>Product - product - object</li>
#                   <li>hResult - ProductService result - ref.hash</li>
#                   <li>ProductInfo - some usefull product attribute values from product - ref.hash</li>
#                   <li>ProductData - additional elements from shop needed for formatting etc.  - ref.hash</li>
#                   </ul> | ref.hash
#========================================================================================
sub OnProductServiceGetProductData {
my ($hParams) = @_;

    my $Product               = $hParams->{'Product'};
    my $ProductID             = $Product->id;
    my $hProductServiceResult = $hParams->{'hResult'};
    my $hProductInfo          = $hParams->{'ProductInfo'};
    my $hShopData             = $hParams->{'ProductData'};
    my $SessionUserID         = $hParams->{'SessionUserID'};
    my $Shop                  = $Product->getSite();
    my $CurrencyID            = $hShopData->{'CurrencyID'};
    my $UserID                = $hParams->{'UserID'};
    my $hTaxCountryData       = $hParams->{'TaxCountryData'};

    # do this only for unity
    return unless $Shop->get('IsUnityRegistered');


    my $TaxMatrix = exists $hTaxCountryData->{'TaxMatrix'} ? $hTaxCountryData->{'TaxMatrix'} : $hShopData->{'TaxMatrix'} ;
    my $TaxModel = $hShopData->{'TaxModel'} ;
    #-- It isn't entirely clear (to me at least) why this "$Shop->get('TaxModel') == 1" check is necessary, but it is used in most places that have anything to do with dt
    #-- See e.g. DE_EPAGES::Product::API::ProductService::getProductData
    #-- It basically says that if the shop tax model is NET, we never want to use the tax model based on the destination country
    #-- So only if the shop tax model is GROSS will there be a potential switch of the tax model to the tax model of the destination country
    if ( exists $hTaxCountryData->{'TaxModel'} && $Shop->get('TaxModel') == 1 ) {
        $TaxModel = $hTaxCountryData->{'TaxModel'};
    }

    my $TaxClass = $Product->getMappedTaxClassByTaxMatrix($TaxMatrix);

    if (defined $UserID && $UserID > 0) {
        $SessionUserID = $UserID;
    }


    # 1st get all Prices and discounts from PriceLists for the product and store to $hAllPrices
    my $hAllPrices = {};
    my $ahPriceLists = $Shop->get('PriceLists');
    foreach my $PriceList (@$ahPriceLists) {
        my $aPrices = $PriceList->getScalePricesByProductID($ProductID, $TaxModel, $TaxClass);
        # reorg prices like discounts
        my $hPrices = {};
        $hPrices->{$_->{'Quantity'}} = $_->{'Price'} foreach (@$aPrices);
        my $hPriceDiscounts = {};
        $hPriceDiscounts->{$_->{'Quantity'}} = $_->{'Discount'} foreach @{ $PriceList->getQuantityDiscountsByProductID($ProductID) };
        $hAllPrices->{$PriceList->alias} = {'Discounts' => $hPriceDiscounts,
                                            'Prices'    => $hPrices};
    }

    # backwards compatibility: add missing entries to prices
    foreach my $KG (keys(%$hAllPrices)){
        my $hPrices = $hAllPrices->{$KG}->{'Prices'};
        my $hDiscounts = $hAllPrices->{$KG}->{'Discounts'};
        # add missing price entry if we have a discount for a quantity which is not existing in the prices
        # this can happen for migrated or older now shops
        foreach my $DiscountQTY ( sort {$a <=> $b}keys(%$hDiscounts)){
            unless(exists $hPrices->{$DiscountQTY} ){
                my $PrevQty;
                foreach my $PriceQty (sort {$a <=> $b} keys(%$hPrices)) {
                    $hPrices->{$DiscountQTY} = $hPrices->{$PrevQty} if $DiscountQTY <= $PriceQty ;
                    $PrevQty = $PriceQty;
                }
            }
        }

        # backwards compatibility add qty 1 to prices if not exists
        my $QtyOne = 1;
        if (ElementOf($QtyOne, keys %$hDiscounts) and not ElementOf($QtyOne, keys(%$hPrices))){
            #-- UNITY-8405
            #-- Accessing a key of a presumed hash, where the variable is undefined, will turn it into a hash, apparently
            #-- So check if it is a hash beforehand
            #-- But we need a price for the quantity of 1? Well, if the list price is undefined ... *shrug*
            $hPrices->{$QtyOne} = $hProductServiceResult->{'priceInfo'}->{'price'}->{'amount'} if ref $hProductServiceResult->{'priceInfo'}->{'price'} eq 'HASH';
        }


        # START build structure and sore data for each customer group
        my $Formatter  = $hShopData->{'Formatter'};
        foreach my $PriceQty (sort {$a <=> $b} keys(%$hPrices)){
            # try to find entry in discounts (think about inheritance)
            my $DiscountQuantityToUse;
            foreach my $Qty (reverse(0 .. $PriceQty)) {
                if ( ElementOf($Qty , keys %$hDiscounts ) ) {
                    $DiscountQuantityToUse = $Qty;
                    last;
                }
            }

            # prepare discount prices
            # UNITY-8479, hPriceSavings must be defined otherwise error with ep-ui
            # When a discount for this quantity cannot be found (life's not fair huh? >.<), assume it to be 0 (after all, what else would it be :D)
            my $DiscountQuantityValue = exists $hDiscounts->{$DiscountQuantityToUse} ? $hDiscounts->{$DiscountQuantityToUse} : 0;
            my $hPriceSavings = undef;
            my %Percent = (
                'percentage' => $DiscountQuantityValue,
                'formatted'  => $Formatter->percentage($DiscountQuantityValue)->{'formatted'},
            );
            $hPriceSavings->{'percent'} = \%Percent;
            $hPriceSavings->{'regularBulkPrice'} = $Formatter->price($hPrices->{$PriceQty}, $CurrencyID, $TaxModel);

            my $Price = $hPrices->{$PriceQty} - (($hPrices->{$PriceQty} * $DiscountQuantityValue) / 100) ;
            $hPrices->{$PriceQty} = $hShopData->{'SiteLocale'}->roundMoney($Price,$CurrencyID);

            my $hPriceInfo = $Formatter->price( $hPrices->{$PriceQty}, $CurrencyID, $TaxModel );
            my $hQuantityInfo = $Formatter->quantity($PriceQty, $hProductInfo->{'OrderUnit'} );
            my $hBasePriceInfo = undef;

            # if the product has a reference price we have to add the basePrice info to the bulk prices
            if(defined $hProductServiceResult->{'priceInfo'}->{'basePrice'}) {
                my $ReferencePrice = $Product->getReferencePrice($CurrencyID, $hPrices->{$PriceQty});
                my $RefPriceFormatted = sprintf '%d %s = %s',
                                        $hProductInfo->{'RefAmount'},
                                        $hProductServiceResult->{'priceInfo'}->{'basePrice'}->{'refQuantity'}->{'unit'},
                                        $hShopData->{'SiteLocale'}->formatMoney($ReferencePrice, $CurrencyID);

                $hBasePriceInfo->{'formatted'} = $RefPriceFormatted;
                $hBasePriceInfo->{'refPrice'} = $Formatter->price( $ReferencePrice, $CurrencyID, $TaxModel );
                $hBasePriceInfo->{'quantity'} = $hProductServiceResult->{'priceInfo'}->{'basePrice'}->{'quantity'};
                $hBasePriceInfo->{'refQuantity'} = $hProductServiceResult->{'priceInfo'}->{'basePrice'}->{'refQuantity'};
            }

            push @{$hAllPrices->{$KG}->{'Final'}}, {
                                'price'         => $hPriceInfo,
                                'quantity'      => $hQuantityInfo,
                                'basePrice'     => $hBasePriceInfo,
                                'priceSavings'  => $hPriceSavings,
                               };


        } # END build structure

    }

    # if we have a session info we use the bulkPriceInfo also for the
    # customer specific session prices
    # without a session bulkPriceInfo stores only the plain bulk prices
    my $SessionCustomerPriceListAlias = UNITY_PRICELIST_PREFIX . $CurrencyID;
    if (defined $SessionUserID) {
        my $aSessionUserCustomers = LoadObject($SessionUserID)->get('Customers');
        if (scalar @$aSessionUserCustomers) {
            my $Customer = $aSessionUserCustomers->[0];
            if( (defined $Customer) and (defined $Customer->get('CustomerGroup')) ){
                my $CustomerGroupPriceList = @{$Customer->get('CustomerGroup')->get('PriceLists')}[0];
                $SessionCustomerPriceListAlias = $CustomerGroupPriceList->alias if defined $CustomerGroupPriceList;
            }
        }
    }

    # collect normal bulk prices or session specific prices
    my $aBulkPrices = [];
    $aBulkPrices = $hAllPrices->{$SessionCustomerPriceListAlias}->{'Final'} // [];
    push @{$hProductServiceResult->{'bulkPriceInfo'}}, @$aBulkPrices;

    # collect customer specific price info (all groups and prices)
    my $ahCustomerSpecificPrices = [];
    foreach my $KG (keys(%$hAllPrices)){
        next if $KG eq  UNITY_PRICELIST_PREFIX . $CurrencyID;
        next unless $hAllPrices->{$KG}->{'Final'};
        
        my $aCustomerGroups = $Shop->folder('PriceLists')->child($KG)->get('CustomerGroups');
        next unless scalar @$aCustomerGroups > 0;
        
        push @$ahCustomerSpecificPrices, { 'customerGroup'   => $KG,
                                           #'customerGroupId' => $Shop->folder('Groups')->child($KG)->guid,
                                           'customerGroupId' => $aCustomerGroups->[0]->guid,
                                           'priceInfo'       => $hAllPrices->{$KG}->{'Final'} };

    }

    push @{$hProductServiceResult->{'customerSpecificPriceInfo'}}, @$ahCustomerSpecificPrices;

}

1;
