#========================================================================================
# package      DE_EPAGES::Zettle::API::ZettleService
# state        public
#----------------------------------------------------------------------------------------
# description  object for calling functions on the ZettleService which serves as a Zettle facade
#========================================================================================
package DE_EPAGES::Zettle::API::ZettleService;

use strict;

use Data::Compare;
use JSON qw ( decode_json );

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( GetLog LogError );
use DE_EPAGES::Core::API::PerlTools qw ( GetCurrentDateTime unique GenerateGUID);
use DE_EPAGES::Core::API::TestParameter  qw( TestObject );
use DE_EPAGES::Object::API::Factory  qw ( ExistsObjectByGUID ExistsObjectByPath LoadObjectByGUID LoadObjectByPath );
use DE_EPAGES::Zettle::API::Constants qw (
    EXPORT_STATUS
    API_IMAGE_FORMAT
    WAIT_INTERVAL_AFTER_DELETE
    MAX_PRODUCTS_PER_EXPORT
    PRODUCT_EXPORT_WAIT_INTERVAL
    MAX_IMAGES_PER_EXPORT
    IMAGES_EXPORT_WAIT_INTERVAL
    API_MAX_PRODUCTS_PER_DAY
    EXPORT_TYPE
    DEFAULT_ZETTLE_WEBHOOKS
    MAX_DB_ERROR_TYPE_LENGTH
    MAX_DB_ERROR_MESSAGE_LENGTH
);
use DE_EPAGES::Zettle::API::ImageHelper qw ( GetGuidImageMap GetImageUploadResults );
use DE_EPAGES::Zettle::API::ProductChangeHelper qw (
    GetModifiedImageProducts
    GetModifiedBasicDataProducts
    GetModifiedPriceProducts
);
use DE_EPAGES::Zettle::API::TokenHelper;
use DE_EPAGES::Zettle::API::Tools qw ( LoadProductByZettleUUID );
use DE_EPAGES::Zettle::API::WebService::AuthService;
use DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateDTO;
use DE_EPAGES::Zettle::API::WebService::DTO::WebhookEventDTO;
use DE_EPAGES::Zettle::API::WebService::ImageService;
use DE_EPAGES::Zettle::API::WebService::InventoryService;
use DE_EPAGES::Zettle::API::WebService::MerchantService;
use DE_EPAGES::Zettle::API::WebService::PusherService;
use DE_EPAGES::Zettle::API::WebService::ProductService;
use DE_EPAGES::Zettle::API::WebService::Requirements::Product qw ( FilterExportableProducts );
use DE_EPAGES::Zettle::API::WebService::RestService qw ( WEB_SERVICE_ERROR_CODE );
use DE_EPAGES::Zettle::API::WebService::VariationOptionHelper qw ( BuildVariantOptions );
use DE_EPAGES::Zettle::API::Logger;
use DE_EPAGES::Zettle::API::Table::ZettleProductExport qw (
    InsertZettleProductExport
    UpdateZettleProductExport
    GetInfoZettleProductExportByStatus
);

# Length of error message field in zettleproductexport DB table
use constant ZETTLE_PRODUCT_EXPORT_ERROR_MESSAGE_LENGTH => 255;

#ERRORDEFINITION
sub ErrorDefinition {
    return {
        'InitialExportError' => 'Initial export failed (GlobalExportUUID=#GlobalExportUUID), ErrorMessage=#ErrorMessage, FailedProducts=#FailedProducts',
        'InvalidOrgUUID' => 'Invalid organizationUuid received in #EventName event: #OrganizationUuid (MessageId: #MessageId)',
    };
}
#ERRORDEFINITION END

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::ZettleService->new($Environment)
#----------------------------------------------------------------------------------------
# description  creates a new ZettleService object
#----------------------------------------------------------------------------------------
# input        $Environment | the Zettle environment object | DE_EPAGES::Zettle::API::Object:ZettleEnvironment
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::ZettleService
#========================================================================================
sub new {
    my $class = shift;
    my ($Environment)  = @_;
    TestObject('ZettleEnvironment', $Environment, 'DE_EPAGES::Zettle::API::Object::ZettleEnvironment');
    
    my $self  = bless {
        'Environment' => $Environment
    }, $class;

    return $self;
}

#========================================================================================
# function     getEnvironment
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Environment = $self->getEnvironment()
#----------------------------------------------------------------------------------------
# description  Returns the Zettle Environment object.
#----------------------------------------------------------------------------------------
# return       $Environment | Zettle Environment | Object
#========================================================================================
sub getEnvironment {
    return shift->{'Environment'};
}

#========================================================================================
# function     connect
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->connect($AuthCode)
#----------------------------------------------------------------------------------------
# description  Use given $AuthCode to retrieve tokens from Zettle OAuth API and create/update zettle environment storing token data.
#               See also <function DE_EPAGES::Zettle::API::WebService::AuthService::getTokensOnAuthCode>.
#========================================================================================
sub connect {
    my $self = shift;
    my ($AuthCode) = @_;

    my $Environment = $self->getEnvironment();

    # Use OAuth service to exchange AuthCode for Access and Refresh tokens
    my $AuthService = DE_EPAGES::Zettle::API::WebService::AuthService->new($Environment);
    my $hTokens = $AuthService->getTokensOnAuthCode($AuthCode);
    
    my $TokenHelper = $self->_createTokenHelper($Environment);
    $TokenHelper->updateTokenData($hTokens);

    my $MerchantService = DE_EPAGES::Zettle::API::WebService::MerchantService->new($Environment);
    # See example in: DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.4.2. Retrieve a Merchant's Account Information)
    my $AccountInfo = $MerchantService->getAccountInfo();
    
    my $ContactEmail = $AccountInfo->{'contactEmail'};
    my $ReceiptEmail = $AccountInfo->{'receiptEmail'};
    my %MerchantInfo = (
        'uuid'     => $AccountInfo->{'uuid'},
        'email'    => $ContactEmail // $ReceiptEmail,
        'country'  => $AccountInfo->{'country'},
        'language' => $AccountInfo->{'language'},
        'currency' => $AccountInfo->{'currency'},
        'taxationType' => $AccountInfo->{'taxationType'},
        'taxationMode' => $AccountInfo->{'taxationMode'},
    );

    eval {
        # If merchant previously connected to Zettle with a different account we cleanup existing export configuration
        my $StoredMerchantUUID = $Environment->getMerchantUUID();
        if(defined($StoredMerchantUUID) && $StoredMerchantUUID ne $MerchantInfo{'uuid'}){
            $Environment->resetExportConfig();
        }
        # Perform necessary actions on reconnection to reach a clean state
        elsif ($self->isInitialExportComplete()) {
            # [STEP] Retrieve location UUIDs from Zettle
            $self->updateAllLocationUUID($Environment);

            # [STEP] Retrieve exportable products created before last export
            my $Shop = $Environment->getSite();
            my $aShopProducts = $Shop->get('Products');
            my ($aExportableProducts) = FilterExportableProducts($aShopProducts);

            my $LastProductExport = $Environment->lastProductExport();
            my @aNewProducts = $self->_getCreatedAndClonedProducts($aExportableProducts, $LastProductExport, { 'IsReconnection' => 1});
            my %hNewProducts = map { $_ => 1 } @aNewProducts;
            my @aOldProducts = grep { not $hNewProducts{$_} } @{$aExportableProducts};

            # [STEP] Enable tracking and export stock levels for previously exported products
            if (scalar @aOldProducts) {
                # Re-enable InventoryTracking
                my $hProductBalanceBulk = $self->_enableInventoryTracking(\@aOldProducts);

                # Export updated stock levels
                $self->_exportStockLevels(\@aOldProducts, $hProductBalanceBulk);
            }
        }

        $Environment->setMerchantInfo(%MerchantInfo);
    };
    # Abort setup if merchant info could not be updated:
    # - disconnect from Zettle (revoke tokens)
    # - delete ZettleEnvironment object
    if( ExistsError() ){
        my $Error = GetError();
        $self->disconnect();
        $Environment->delete();
        Error($Error);
    };
}

#========================================================================================
# function     disconnect
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->disconnect($RemoveExportedProductsFromZettle,$MarkForDeletion)
#----------------------------------------------------------------------------------------
# input        $RemoveExportedProductsFromZettle | Optional. If true, the products will be deleted from zettle | boolean
# input        $MarkForDeletion | Optional. Determines whether ZettleEnvironment will be marked to be deleted by batch job (default: TRUE) | boolean
#----------------------------------------------------------------------------------------
# description  Disconnect from Zettle API and remove token data from zettle environment
#========================================================================================
sub disconnect {
    my $self = shift;
    my ($RemoveExportedProductsFromZettle,$MarkForDeletion) = @_;

    $MarkForDeletion //= 1;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();

    $self->_removeEnvironmentSubscription($Environment);
                  
    # Disable inventory tracking for all tracked shop products
    my @aTrackedProducts = grep { $_->get('ZettleTrackingEnabled') } @{ $Shop->get('Products') };
    eval {
        $self->_disableInventoryTracking(\@aTrackedProducts);
    };
    if(ExistsError()){ LogError(GetError()); }

    #Remove products from zettle
    if($RemoveExportedProductsFromZettle){
        eval {
            my $aShopProducts = $Shop->get('Products');

            #Remove all the product from the shop in Zettle catalog
            $self->deleteProducts($aShopProducts);

            #Remove ZettleUUID from the deleted products
            $self->removeProductsZettleUUID($aShopProducts);
        };
        if (ExistsError()) {
            #Even if there are errors when deleting, we ignore them and continue with the disconnection
            my $Error = GetError();
            LogError($Error);
        }
    }

    # Delete connection
    my $AuthService = DE_EPAGES::Zettle::API::WebService::AuthService->new($Environment);
    eval {
        $AuthService->disconnect();
    };
    if(ExistsError()){ LogError(GetError()); }

    my $TokenHelper = $self->_getTokenHelper();
    $TokenHelper->deleteTokenData();

    # Configure ZettleEnvironment so that batch job proceeds with deletion afterwards if MarkForDeletion is true
    $Environment->markForDeletion($MarkForDeletion);
}

#========================================================================================
# function     isTokenAcquired
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->isTokenAcquired()
#----------------------------------------------------------------------------------------
# description  Check if tokens were retrieved from Zettle
#               See <function DE_EPAGES::Zettle::API::TokenHelper::isTokenAcquired>
#----------------------------------------------------------------------------------------
# return        TRUE if tokens present in ZettleEnvironment, false otherwise
#========================================================================================
sub isTokenAcquired {
    my $self = shift;
    
    my $TokenHelper = $self->_getTokenHelper();
    return $TokenHelper->isTokenAcquired();
}

#========================================================================================
# function     initialExport
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hExportResult = $self->initialExport()
#----------------------------------------------------------------------------------------
# description  Run an initial product export which will send to Zettle all products that match criteria,
#               and update InitialExportStatus in environment.
#               See <function DE_EPAGES::Zettle::API::WebService::ProductService::exportProducts>.
#               See <function DE_EPAGES::Zettle::API::WebService::Requirements::Product::TestRequirements>.
#----------------------------------------------------------------------------------------
# input       $ReplaceZettleLibrary | Determines whether all products will be removed from Zettle before export | boolean
# input       $IncludePrices | Determines if prices will be included in product export | boolean
#----------------------------------------------------------------------------------------
# return       $hExportResult | API response for product export and list of filtered and failed products, with following keys:
#               <ul>
#                   <li>GlobalExportUUID - General identifier generated - string
#                   <li>Status - Status defined from all the batch export. Any of the values defined
#                       in constant EXPORT_STATUS (see <package DE_EPAGES::Zettle::API::Constants>) - int
#                   <li>CheckFailed - optional. List of products (ProductError) that weren't exported for not meeting export requirements - ref.array.DE_EPAGES::Zettle::API::WebService::Requirements::ProductError
#                   <li>ExportFailed - optional. IDs of products that weren't exported due to errors on Zettle processing - ref.array.int
#                   <li>ErrorMessage - optional. A general error message - string
#               </ul> | ref.hash
#========================================================================================
sub initialExport {
    my $self = shift;
    my ($ReplaceZettleLibrary, $IncludePrices) = @_;
    
    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
    
    my $MethodName = (caller(0))[3];
    GetLog->debug("$MethodName: Shop=".$Shop->alias.", ReplaceZettleLibrary=$ReplaceZettleLibrary, IncludePrices=$IncludePrices");

    eval {
        # [STEP] Initialize status
        $Environment->initialExportStatus(EXPORT_STATUS->{RUNNING});

        if ($ReplaceZettleLibrary) {
            my $hDeleteResult = $ProductService->deleteAllProducts();
            my $aFailedDeletion = $hDeleteResult->{'failed'};
            if (scalar @{$aFailedDeletion} > 0) {
                GetLog->warn('Could not delete some products from Zettle library - Zettle UUIDs: '.join(', ', @{$aFailedDeletion}));
            }
            # PATCH: we must wait a few seconds before starting initial export process so that Zettle correctly shows all products in dashboard
            sleep(WAIT_INTERVAL_AFTER_DELETE);
        } else {
            # [STEP] ZettleUUIDs: use product info retrieved from Zettle to fill in product ZettleUUID attribute for previously exported products/variations
            my $aProductResponseDTO = $ProductService->getAllProducts();
            $self->_updateShopProductsUUID($aProductResponseDTO);
        }

        # [STEP] Retrieve all exportable products
        my $aShopProducts = $Shop->get('Products');
        my ($aExportableProducts, $hFilteredProducts) = FilterExportableProducts($aShopProducts);
        GetLog->debug("FilterExportableProducts returned: \n".
                        "- aExportableProducts: " . join(', ', map { $_->get('RelativePath') } @{$aExportableProducts}) . "\n".
                        "- filteredProducts: " . join(', ', keys %{$hFilteredProducts}));

        # [STEP] ProductImages: export all product images and retrieve imageUploadResults to be used in product export.
        my $hImageUploadResults = $self->_exportImages($aExportableProducts, API_IMAGE_FORMAT);

        # [STEP] ProductData: export fields of interest (which will include image lookupKeys from $hImageUploadResults)
        my $hExportResult = $self->_exportProductsInBatches($aExportableProducts, EXPORT_TYPE->{INITIAL_EXPORT}, $IncludePrices, $hImageUploadResults);
        
        # Combine failed products in exportProducts() and products filtered in FilterExportableProducts()
        push(@{$hExportResult->{'CheckFailed'}}, values %{$hFilteredProducts});

        my $InitialExportUUID = $hExportResult->{GlobalExportUUID};
        $Environment->setupInitialExport($InitialExportUUID, $IncludePrices);

        my $Status = $hExportResult->{'Status'};
        $Environment->initialExportStatus($Status);

        # Check whether export failed
        if ($Status == EXPORT_STATUS->{FAILED}) {
            my $ahBatchExportInfo = $hExportResult->{'BatchExportInfo'};
            my @aFailedBatches = grep { exists $_->{'Error'} } @{$ahBatchExportInfo};

            my @aErrorMessages = ();
            my $hFailedProductIDs = {};
            foreach my $hBatchExportInfo (@aFailedBatches) {
                push(@aErrorMessages, $hBatchExportInfo->{'Error'}->message);
                $hFailedProductIDs->{$_} = 1 foreach @{$hBatchExportInfo->{'ExportFailed'}};
            }
            
            my $GlobalErrorMessage = $hExportResult->{'ErrorMessage'} // '';
            my $ErrorMessage = "[$GlobalErrorMessage ".join(', ', @aErrorMessages).']';
            my $ExportFailedProducts = (scalar keys %{$hFailedProductIDs}) ? join(',', sort keys %{$hFailedProductIDs}) : 'UNKNOWN';
            Error('InitialExportError', {
                'GlobalExportUUID' => $InitialExportUUID,
                'ErrorMessage' => $ErrorMessage,
                'FailedProducts' => '['.$ExportFailedProducts.']'
            });
        }

        # [STEP] Retrieve location UUIDs from Zettle
        $self->updateAllLocationUUID($Environment);

        if (scalar @{$aExportableProducts}) {
            # [STEP] Enable InventoryTracking for exported products
            $self->_enableInventoryTracking($aExportableProducts);

            # [STEP] Export stock levels for tracked products
            $self->_exportStockLevels($aExportableProducts);
        }

        #TODO: consider possible errors when enabling tracking or exporting stock levels and add info to 'failed' items
        return $hExportResult;
    };
    if (ExistsError()) {
        my $Error = GetError();
        $Environment->initialExportStatus(EXPORT_STATUS->{FAILED});
        Error($Error);
    }
}

#========================================================================================
# function     _exportProductsInBatches
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hExportResult = $self->_exportProductsInBatches($aProducts, $ExportType, $IncludePrices, $hImageUploadResults);
#----------------------------------------------------------------------------------------
# description  Export all product/price information from the shop to Zettle in batches.
#               See <function DE_EPAGES::Zettle::API::WebService::ProductService::exportProducts>.
#               See <function DE_EPAGES::Zettle::API::WebService::ProductService::exportPrices>.
#----------------------------------------------------------------------------------------
# input        $aProducts | Products to be exported | ref.array.object
# input        $ExportType | Type of export (INITIAL_EXPORT, PRODUCT_UPDATE, PRICE_UPDATE),
#                             see EXPORT_TYPE values in <package DE_EPAGES::Shop::API::Constants> | integer
# input        $IncludePrices | Determines whether price information will be exported
#                                (applies only if $ExportType is not EXPORT_TYPE->{PRICE_UPDATE}) | boolean
# input        $hImageUploadResults | Image upload results (result of _exportImages() call) | ref.hash
#----------------------------------------------------------------------------------------
# return       $hExportResult | with following keys:
#               <ul>
#                   <li>GlobalExportUUID - General random identifier generated for this export (UUIDv1) - string
#                   <li>Status - General status calculated from the combination of all batch exports statuses.
#                       Refer to values of constant EXPORT_STATUS (see <package DE_EPAGES::Zettle::API::Constants>) - int
#                   <li>BatchExportInfo - detailed information of each batch export. A hash with the following keys:
#                       <ul>
#                           <li>BatchNumber - position of batch in the sequence of export - int
#                           <li>Status - latest status retrieved from Zettle for this batch - int
#                           <li>BatchUUID - optional. export UUID for this batch (returned by Zettle if no error occurred) - int
#                           <li>Error - optional. Error object in case an error occurred - DE_EPAGES::Core::API::Error
#                           <li>CheckFailed - optional. List of products (ProductError) that weren't exported for not meeting export requirements - ref.array.DE_EPAGES::Zettle::API::WebService::Requirements::ProductError
#                           <li>ExportFailed - optional. IDs of products that weren't exported due to errors on Zettle processing - ref.array.int
#                           <li>ErrorMessage - optional. General description of global error - string
#                       </ul> - ref.array.hash
#               </ul> | ref.hash
#========================================================================================
sub _exportProductsInBatches {
    my $self = shift;
    my ($aProducts, $ExportType, $IncludePrices, $hImageUploadResults) = @_;
    
    # A random unique identifier will be generated and associated to each export job.
    # Note that this is not the UUID used by Zettle to identify a running export, but an internal ID for us.
    my $GlobalExportUUID = GenerateGUID();

    my $Environment = $self->getEnvironment();
    my $ShopID = $Environment->getSite()->id;
    
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);

    my $ProductCount = scalar @{$aProducts};
    # [STEP] Check daily export limit
    if (_ExportExceedsDailyLimit($ShopID, $aProducts)) {
        return {
            'GlobalExportUUID' => 'undef',
            'Status' => EXPORT_STATUS->{FAILED},
            'ErrorMessage' => 'Products to export exceeds limit: '.API_MAX_PRODUCTS_PER_DAY.' items per 24hrs'
        };
    }
    my $hGlobalExportResult = {
        'GlobalExportUUID' => $GlobalExportUUID
    };
    my $ahBatchExportInfo = [];
    my $BatchCounter = 0;
    for (my $StartIndex=0; $StartIndex<$ProductCount; $StartIndex+=MAX_PRODUCTS_PER_EXPORT) {
        my $BatchExportUUID;
        my ($BatchError, $ErrorType, $ErrorMessage);
        my $ItemCount;
        my $BatchStatus = EXPORT_STATUS->{RUNNING};
        my $hBatchExportInfo = {
            'BatchNumber' => ++$BatchCounter,
            'Status' => $BatchStatus
        };

        # Collect a batch of products
        my $EndIndex = ($StartIndex + MAX_PRODUCTS_PER_EXPORT >= $ProductCount) ? $ProductCount-1 : $StartIndex + MAX_PRODUCTS_PER_EXPORT -1;
        my @aBatchProducts = @{$aProducts}[$StartIndex..$EndIndex];
        $ItemCount = $EndIndex-$StartIndex+1;
        eval {
            # Register new batch export entry in DB
            InsertZettleProductExport({
                'SiteID' => $ShopID,
                'GlobalExportUUID' => $GlobalExportUUID,
                'BatchNumber' => $BatchCounter,
                'ExportType' => $ExportType,
                'ItemCount' => $ItemCount,
                'BatchExportStatus' => EXPORT_STATUS->{RUNNING},
            });
            
            my $hPartialExportResult;
            if ($ExportType == EXPORT_TYPE->{PRICE_UPDATE}) {
                $hPartialExportResult = $ProductService->exportPrices(\@aBatchProducts);
            } else {
                $hPartialExportResult = $ProductService->exportProducts(\@aBatchProducts, $IncludePrices, $hImageUploadResults);
            }
            
            my $hResponse = $hPartialExportResult->{'response'};
            $BatchExportUUID = $hResponse->{'uuid'};
            $ItemCount = $hResponse->{'items'};

            # Retrieve updated Status
            my $State = $hResponse->{'state'};
            $BatchStatus = $self->_getStatusFromState($State);

            $hBatchExportInfo->{'CheckFailed'} = $hPartialExportResult->{'failed'};

            # Do not overload Zettle product service (avoid IMPORT_ALREADY_RUNNING error)
            sleep(PRODUCT_EXPORT_WAIT_INTERVAL);
        };
        if (ExistsError()) {
            $BatchError = GetError();
            $BatchStatus = EXPORT_STATUS->{FAILED};
            LogError($BatchError);
            my $ErrorVars = $BatchError->vars();
            if ($BatchError->code eq WEB_SERVICE_ERROR_CODE && defined $ErrorVars->{'Response'}) {
                # Zettle API error (see list of errors in DE_EPAGES/Zettle/Documents/API/ZettleAPI-Errors.md)
                my $hResponse = decode_json($ErrorVars->{'Response'});
                $ErrorType = substr($hResponse->{'errorType'}, 0, MAX_DB_ERROR_TYPE_LENGTH) if defined $hResponse->{'errorType'};
                $ErrorMessage = substr($hResponse->{'developerMessage'}, 0, MAX_DB_ERROR_MESSAGE_LENGTH) if defined $hResponse->{'developerMessage'};
                # Compile IDs of all products in batch (when export fails no product is processed by Zettle)
                GetLog->debug("_exportProductsInBatches: ExportFailed aBatchProducts=".join(', ', map { $_->id } @aBatchProducts));
                my @aSortedIds = sort( map { $_->id } @aBatchProducts);
                $hBatchExportInfo->{'ExportFailed'} = \@aSortedIds;
                #my $HttpErrorCode = $ErrorVars->{'StatusCode'};
                #my $aViolations = defined $hResponse->{'violations'} ? $hResponse->{'violations'} : [];
            } else {
                $ErrorType = substr($BatchError->code, 0, MAX_DB_ERROR_TYPE_LENGTH);
                $ErrorMessage = substr($BatchError->message, 0, MAX_DB_ERROR_MESSAGE_LENGTH);
            }
        }

        # Update batch export entry in DB
        my $hUpdateZettleProductExport = { 'BatchExportStatus' => $BatchStatus };
        $hBatchExportInfo->{'Status'} = $BatchStatus;      
        if (defined $BatchError) {
            $hBatchExportInfo->{'Error'} = $BatchError;
            $hUpdateZettleProductExport->{'ErrorType'} = $ErrorType;
            $hUpdateZettleProductExport->{'ErrorMessage'} = $ErrorMessage;
        } else {
            # BatchExportUUID is only available in Zettle response if there is no error
            $hBatchExportInfo->{'BatchUUID'} = $BatchExportUUID;
            $hUpdateZettleProductExport->{'BatchExportUUID'} = $BatchExportUUID;
            # Initial value for ItemCount is a rough estimate since it also counts subproducts. Now we update to precise value with Zettle info
            $hUpdateZettleProductExport->{'ItemCount'} = $ItemCount;
        }
        UpdateZettleProductExport($GlobalExportUUID, $BatchCounter, $hUpdateZettleProductExport);

        push (@{$ahBatchExportInfo}, $hBatchExportInfo);
    }

    $hGlobalExportResult->{'Status'} = $self->getGlobalExportStatus($GlobalExportUUID);
    $hGlobalExportResult->{'BatchExportInfo'} = $ahBatchExportInfo;

    return $hGlobalExportResult;
}

#========================================================================================
# function     getGlobalExportStatus
# state        public
#----------------------------------------------------------------------------------------
# syntax       $GlobalStatus = $self->getGlobalExportStatus($GlobalExportUUID)
#----------------------------------------------------------------------------------------
# description  Return global status as the combination of all batches statuses.
#               Updated status for running batches will be retrieved from Zettle.
#               Global status will be calculated as follows:
#               - If there is any RUNNING batch -> 1 (RUNNING).
#               - If there is no RUNNING batch and there is at least one FAILED batch -> -1 (FAILED).
#               - If all batches are in COMPLETED status -> 2 (COMPLETED).
#----------------------------------------------------------------------------------------
# input        $GlobalExportUUID | Global export UUID - varchar(36) | string
#----------------------------------------------------------------------------------------
# return       $GlobalStatus | See EXPORT_STATUS values in <package DE_EPAGES::Zettle::API::Constants> | int
#========================================================================================
sub getGlobalExportStatus {
    my $self = shift;
    my ($GlobalExportUUID) = @_;

    my $Environment = $self->getEnvironment();
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);

    my $IsGlobalExportComplete = 1;

    # Collect RUNNING exports and query Zettle to retrieve update state
    my $ahPendingBatches = GetInfoZettleProductExportByStatus($GlobalExportUUID, EXPORT_STATUS->{RUNNING});
    foreach my $hBatchExport (@{$ahPendingBatches}) {
        my $BatchNumber = $hBatchExport->{'BatchNumber'};
        my $BatchExportUUID = $hBatchExport->{'BatchExportUUID'};

        # Some export operations fail early and we won't get a UUID for such export
        next unless defined $BatchExportUUID;

        my $hExportState = $ProductService->getBulkExportState($BatchExportUUID);

        # Convert export state ("IMPORTING", "FINISHED_SUCCESS", "FINISHED_FAILED") to batch status (see EXPORT_STATUS)
        my $BatchState = $hExportState->{'state'};
        my $BatchStatus = $self->_getStatusFromState($BatchState);
        $IsGlobalExportComplete = 0 if ($BatchStatus == EXPORT_STATUS->{RUNNING});
        
        my $hUpdateExport = {
            'BatchExportStatus' => $BatchStatus
        };
        if ($BatchStatus == EXPORT_STATUS->{FAILED}) {
            $hUpdateExport->{'ErrorType'} = 'API_IMPORT_FAILURE';
            $hUpdateExport->{'ErrorMessage'} = substr($hExportState->{'errorMessage'}, ZETTLE_PRODUCT_EXPORT_ERROR_MESSAGE_LENGTH);
        }

        # Update status in database
        UpdateZettleProductExport($GlobalExportUUID, $BatchNumber, $hUpdateExport);
    }

    # When batch export is generated, initial status is RUNNING
    # Then status can either be updated to FAILED or COMPLETE
    if (not $IsGlobalExportComplete) {
        return EXPORT_STATUS->{RUNNING};
    }
    # Check if there is any failed batch
    my $ahFailedBatches = GetInfoZettleProductExportByStatus($GlobalExportUUID, EXPORT_STATUS->{FAILED});
    if (scalar @{$ahFailedBatches} > 0) {
        return EXPORT_STATUS->{FAILED};
    }

    return EXPORT_STATUS->{COMPLETED};
}

#========================================================================================
# function     isInitialExportComplete
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->isInitialExportComplete()
#----------------------------------------------------------------------------------------
# description  Check if initial product export has been completed
#----------------------------------------------------------------------------------------
# return        TRUE if completed, false otherwise
#========================================================================================
sub isInitialExportComplete {
    my $self = shift;
    my $InitialExportStatus = $self->getInitialExportStatus();
    return ($InitialExportStatus == EXPORT_STATUS->{COMPLETED}) ? 1 : 0;
}

#========================================================================================
# function     getInitialExportStatus
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->getInitialExportStatus()
#----------------------------------------------------------------------------------------
# description  Retrieve updated initial export status querying Zettle if necessary (if export was RUNNING), updating environment export status if status changed.
#----------------------------------------------------------------------------------------
# return       $ExportStatus | See EXPORT_STATUS values in <package DE_EPAGES::Zettle::API::Constants> | integer
#========================================================================================
sub getInitialExportStatus {
    my $self = shift;
    my $Environment = $self->getEnvironment();
    my $InitialExportStatus = $Environment->initialExportStatus();

    # If last known status was 'running' we query Zettle for updated status
    if ($InitialExportStatus == EXPORT_STATUS->{RUNNING}) {
        my $InitialExportUUID = $Environment->initialExportUUID();

        # This prevents an error the first time merchant runs initialExport
        # (product export begins but InitialExportUUID is undefined for a short time span - see initialExport() method)
        return $InitialExportStatus unless defined $InitialExportUUID;

        $InitialExportStatus = $self->getGlobalExportStatus($InitialExportUUID);

        # Update export status in environment if status changed
        $Environment->initialExportStatus($InitialExportStatus) unless $InitialExportStatus == EXPORT_STATUS->{RUNNING};
    }

    return $InitialExportStatus;
}

#========================================================================================
# function     exportProductUpdates
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hResult = $self->exportProductUpdates($IncludePrices);
#----------------------------------------------------------------------------------------
# description  Export product basic information to Zettle for products modified after previous export.
#               Also export products created and products deleted after LastProductExport.
#               N.B. Previous product export datetime is stored in ZettleEnvironment 'LastProductExport' field.
#----------------------------------------------------------------------------------------
# return       $hResult | Result of exporting changes to Zettle, with following keys:
#               <ul>
#                   <li>GlobalExportUUID - General random identifier generated for this export (UUIDv1) - string
#                   <li>Status - General status calculated from the combination of all batch exports statuses.
#                       Refer to values of constant EXPORT_STATUS (see <package DE_EPAGES::Zettle::API::Constants>) - int
#                   <li>RequirementErrors - list of products that weren't exported due to API requirement errors - ref.array.DE_EPAGES::Zettle::API::WebService::Requirements::ProductError
#                   <li>DeletedUUIDs - list of UUIDs of products deleted from Zettle - ref.array.integer
#                   <li>DeleteErrors - list of product IDs of failed deletions - ref.array.integer
#                   <li>UpdateCount - number of product changes exported to Zettle - integer
#                   <li>BatchExportInfo - detailed information of each batch export. A hash with the following keys:
#                       <ul>
#                           <li>BatchNumber - position of batch in the sequence of export - int
#                           <li>Status - latest status retrieved from Zettle for this batch - int
#                           <li>BatchUUID - optional. export UUID for this batch (returned by Zettle if no error occurred) - int
#                           <li>Error - optional. Error object in case an error occurred - DE_EPAGES::Core::API::Error
#                           <li>CheckFailed - optional. List of products (ProductError) that weren't exported for not meeting export requirements - ref.array.DE_EPAGES::Zettle::API::WebService::Requirements::ProductError
#                           <li>ExportFailed - optional. IDs of products that weren't exported due to errors on Zettle processing - ref.array.int
#                       </ul> - ref.array.hash
#               </ul>| ref.hash
#========================================================================================
sub exportProductUpdates {
    my $self = shift;
    my ($IncludePrices) = @_;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();
    my $LastProductExport = $Environment->lastProductExport();
    
    my $hResult = {};

    # [STEP] Send latest product deletions to Zettle
    my $hDeleteResult = $self->_exportLastProductDeletions($LastProductExport);
    $hResult->{'DeletedUUIDs'} = $hDeleteResult->{'deleted'};
    $hResult->{'DeleteErrors'} = $hDeleteResult->{'failed'};

    # [STEP] Retrieve shop products created and updated since last export
    # Retrieve products whose image was updated
    my $aUnfilteredImageProducts = GetModifiedImageProducts($Shop, $LastProductExport);
    my ($aProductsModifiedImage, $hFilteredImageProducts) = FilterExportableProducts($aUnfilteredImageProducts);
    # Retrieve products whose basic Zettle-related fields were updated (also includes products cloned or created)
    my $aUnfilteredBasicDataProducts = GetModifiedBasicDataProducts($Shop, $LastProductExport);
    my ($aProductsBasicDataUpdated, $hFilteredBasicDataProducts) = FilterExportableProducts($aUnfilteredBasicDataProducts);
    
    # Basic data shall be exported for newly created products, for products whose basic Zettle-related fields were modified and for products whose image was modified
    my @aDataProducts = unique(@{$aProductsModifiedImage}, @{$aProductsBasicDataUpdated});
    
    # Combine filtered products from ModifiedImage and BasicDataUpdated
    my $hFilteredProducts = $hFilteredImageProducts;
    foreach my $ProductId (keys %{$hFilteredBasicDataProducts}) {
        # Skip products already considered as filtered
        next if exists $hFilteredProducts->{$ProductId};
        $hFilteredProducts->{$ProductId} = $hFilteredBasicDataProducts->{$ProductId};
    }
    $hResult->{'RequirementErrors'} = [ values %{$hFilteredProducts} ];

    # [STEP] Partial export should only export modified images (unlike imagesInitialExport which must export all product images).
    my $hImageUploadResults = $self->_exportImages($aProductsModifiedImage, API_IMAGE_FORMAT);

    # [STEP] Export product basic information
    if (scalar @aDataProducts > 0) {
        my $hExportResult = $self->_exportProductsInBatches(\@aDataProducts, EXPORT_TYPE->{PRODUCT_UPDATE}, $IncludePrices, $hImageUploadResults);

        foreach my $Key (qw( GlobalExportUUID Status BatchExportInfo ErrorMessage)) {
            $hResult->{$Key} = $hExportResult->{$Key} if exists $hResult->{$Key};
        }

        # Retrieve products created/cloned since last export
        my @aNewProducts = $self->_getCreatedAndClonedProducts($aProductsBasicDataUpdated, $LastProductExport);
        if (scalar @aNewProducts) {
            # [STEP] Enable tracking
            $self->_enableInventoryTracking(\@aNewProducts);
            
            # [STEP] Export stock levels
            $self->_exportStockLevels(\@aNewProducts);
        }
    }

    $hResult->{'UpdateCount'} = scalar @aDataProducts;

    return $hResult;
}

#========================================================================================
# function     _getCreatedAndClonedProducts
# state        private
#----------------------------------------------------------------------------------------
# syntax       my @NewProducts = $self->_getCreatedAndClonedProducts($aProducts, $DateTime);
# syntax       my @NewProducts = $self->_getCreatedAndClonedProducts($aProducts, $DateTime, $hOptions);
#----------------------------------------------------------------------------------------
# description  Retrieve products from $aProducts created after the given date.
#               Both newly created and cloned products will be returned.
#               This method requires exportProducts() method to have been run for $aProducts
#               in order for cloned products to have a value set for ZettleUUID.
#----------------------------------------------------------------------------------------
# input        $aProducts   | List of products | ref.array.DE_EPAGES::Product::API::Object::Product
# input        $DateTime    | Time reference to search for new products | Datetime
# input        $hOptions    | Optional. Options with the following keys:
#                              <ul>
#                                  <li>IsReconnection - Indicates that a reconnection operation is undergoing (ignore cloned products) - bool
#                              </ul>
#----------------------------------------------------------------------------------------
# return       @NewProducts | List of products created (and cloned) after given date | array.DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub _getCreatedAndClonedProducts {
    my $self = shift;
    my ($aProducts, $DateTime, $hOptions) = @_;
    $hOptions //= {};

    return () unless scalar @{$aProducts};

    # Retrieve products created after given date
    my @aCreatedProducts = grep { DateTime->compare($DateTime, $_->get('CreationDate')) < 0 } @{$aProducts};

    # For reconnection operations we skip cloned products "calculation" and return only created products.
    # Since upon disconnection ZettleTrackingEnabled is unset for all products, "calculation" would yield all shop products.
    if ($hOptions->{'IsReconnection'}) {
        return @aCreatedProducts;
    }

    # Cloned products share CreationDate with original product hence they're not included in @aCreatedProducts.
    # We identify exported clones by filtering modified products which:
    # - have ZettleUUID defined, which means product was previously exported to Zettle.
    # - don't have ZettleTrackingEnabled set, meaning InventoryTracking was not "STARTED".
    my @aExportedClones = grep { $_->get('ZettleTrackingEnabled') == 0 }
                          grep { defined $_->get('ZettleUUID') }
                          @{ $aProducts };

    return unique(@aCreatedProducts, @aExportedClones);
}

#========================================================================================
# function     exportPriceUpdates
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hExportResult = $self->exportPriceUpdates();
# syntax       $hExportResult = $self->exportPriceUpdates($ExportAllPrices);
#----------------------------------------------------------------------------------------
# description  Export product prices to Zettle:
#               <ul>
#                   <li>if $ExportAllPrices is TRUE, export prices for all exportable products (see <function DE_EPAGES::Zettle::API::WebService::Requirements::Product::FilterExportableProducts>).
#                   <li>if $ExportAllPrices is FALSE or undef, only export prices modified after last price export.
#               </ul>
#----------------------------------------------------------------------------------------
# input        $ExportAllPrices   | Optional. Determines whether to export prices for all products or only modified prices | boolean
#----------------------------------------------------------------------------------------
# return       $hResult | Result of exporting changes to Zettle, with following keys:
#               <ul>
#                   <li>GlobalExportUUID - General random identifier generated for this export (UUIDv1) - string
#                   <li>Status - General status calculated from the combination of all batch exports statuses.
#                       Refer to values of constant EXPORT_STATUS (see <package DE_EPAGES::Zettle::API::Constants>) - int
#                   <li>RequirementErrors - list of products that weren't exported due to API requirement errors - ref.array.DE_EPAGES::Zettle::API::WebService::Requirements::ProductError
#                   <li>UpdateCount - number of product changes exported to Zettle - integer
#                   <li>BatchExportInfo - detailed information of each batch export. A hash with the following keys:
#                       <ul>
#                           <li>BatchNumber - position of batch in the sequence of export - int
#                           <li>Status - latest status retrieved from Zettle for this batch - int
#                           <li>BatchUUID - optional. export UUID for this batch (returned by Zettle if no error occurred) - int
#                           <li>Error - optional. Error object in case an error occurred - DE_EPAGES::Core::API::Error
#                           <li>CheckFailed - optional. List of products (ProductError) that weren't exported for not meeting export requirements - ref.array.DE_EPAGES::Zettle::API::WebService::Requirements::ProductError
#                           <li>ExportFailed - optional. IDs of products that weren't exported due to errors on Zettle processing - ref.array.int
#                       </ul> - ref.array.hash
#               </ul>| ref.hash
#========================================================================================
sub exportPriceUpdates {
    my $self = shift;
    my ($ExportAllPrices) = @_;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();
    
    my $aUnfilteredProducts;
    if (defined $ExportAllPrices && $ExportAllPrices == 1) {
        $aUnfilteredProducts = $Shop->get('Products');
    } else {
        my $LastPriceExport = $Environment->lastPriceExport();
        # Export prices for all products if no previous price export was executed,
        # otherwise export only products whose price was modified
        $aUnfilteredProducts = defined $LastPriceExport
                            ? GetModifiedPriceProducts($Shop, $LastPriceExport)
                            : $Shop->get('Products');
    }
    my ($aExportableProducts, $hFilteredProducts) = FilterExportableProducts($aUnfilteredProducts);

    my $hResult = {};
    push(@{$hResult->{'RequirementErrors'}}, values %{$hFilteredProducts});
    
    if (scalar @{$aExportableProducts} > 0) {
        my $hExportResult = $self->_exportProductsInBatches($aExportableProducts, EXPORT_TYPE->{PRICE_UPDATE});
        foreach my $Key (qw( GlobalExportUUID Status BatchExportInfo ErrorMessage)) {
            $hResult->{$Key} = $hExportResult->{$Key} if exists $hResult->{$Key};
        }
    }
    
    $hResult->{'UpdateCount'} = scalar @{$aExportableProducts};

    return $hResult;

}

#========================================================================================
# function     setupStockWebhook
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->setupStockWebhook($Destination, $ContactEmail, $aEventNames);
#----------------------------------------------------------------------------------------
# description  Subscribe to inventory events to receive stock level updates via Zettle webhooks.
#               See also: <function DE_EPAGES::Zettle::API::ZettleService::incomingStockUpdate>.
#----------------------------------------------------------------------------------------
# input        $Destination  | string
# input        $ContactEmail | string
# input        $aEventNames  | optional. Array of Events we want to subscribe.
#                               See DE_EPAGES::Zettle::API:WebService:PusherService::%SUPORTED_EVENTS.
#                               If null, the subscription will be made for these events:
#                               <ul>
#                                   <li>InventoryBalanceChanged</li>
#                                   <li>InventoryTrackingStarted</li>
#                                   <li>InventoryTrackingStopped</li>
#                               </ul> | ref.array.string
#----------------------------------------------------------------------------------------
# notes        After subscription is created a POST request is sent to $Destination when subscribed event triggers.
#               If Zettle request fails (is not correctly processed by ePages) an email is sent to $ContactEmail.
#========================================================================================
sub setupStockWebhook {
    my $self = shift;
    my ($Destination, $ContactEmail, $aEventNames) = @_;

    my $Environment = $self->getEnvironment();

    if(!defined($aEventNames) || scalar(@$aEventNames) == 0){
        # Inventory-related events
        $aEventNames = DEFAULT_ZETTLE_WEBHOOKS;
    }
    
    my $MethodName = (caller(0))[3];

    my $PusherService = DE_EPAGES::Zettle::API::WebService::PusherService->new($Environment);
    # ShopDomain is used to specifically filter subscriptions generated from our shop
    my $ShopDomain = $Environment->get('Site')->get('ProtocolAndServerSSL');

    # Get subscriptions for our shop registered in Zettle
    my $aSubscriptionResponseDTO = $PusherService->getAllSubscriptions($ShopDomain);
    my $activeSubscriptionsCounter = scalar(@$aSubscriptionResponseDTO);

    my $SubscriptionResponseDTO;
    my $deleteSubscriptionsAndCreateANewOne = 0;
    if ($activeSubscriptionsCounter == 1) {
        my $CurrentSubscriptionUUID = $Environment->getInventorySubscriptionUUID();
        $SubscriptionResponseDTO = $aSubscriptionResponseDTO->[0];
        if (!defined($CurrentSubscriptionUUID)) {
            # We don't have a SubscriptionUUID
            $deleteSubscriptionsAndCreateANewOne = 1;
        } elsif ($CurrentSubscriptionUUID ne $SubscriptionResponseDTO->uuid
                || !Data::Compare::Compare($aEventNames, $SubscriptionResponseDTO->eventNames)) {
            # We have a SubscriptionUUID but doesn't match Zettle's or event names don't match
            $deleteSubscriptionsAndCreateANewOne = 1;
        }
    } else {
        # Zettle has no subscription or has multiple subscriptions
        $deleteSubscriptionsAndCreateANewOne = 1;
    }

    if ($deleteSubscriptionsAndCreateANewOne) {
        # Delete existing subscriptions in Zettle, if any
        foreach my $SubscriptionToRemove (@$aSubscriptionResponseDTO) {
            $PusherService->deleteSubscription($SubscriptionToRemove->uuid);
        }
        
        # Create a new subscription
        GetLog->debug("$MethodName: registering stock webhooks - ".join(',', @{$aEventNames}));
        $SubscriptionResponseDTO = $PusherService->createSubscription($Destination, $ContactEmail, $aEventNames);
        # Update InventorySubscriptionUUID and InventorySigningKey
        my $Uuid = $SubscriptionResponseDTO->uuid();
        my $SigningKey = $SubscriptionResponseDTO->signingKey();
        # Update InventorySubscriptionUUID and InventorySigningKey
        $Environment->setupInventorySubscription($Uuid, $SigningKey);
    }
}

#========================================================================================
# function     incomingStockUpdate
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->incomingStockUpdate($hEventContent);
#----------------------------------------------------------------------------------------
# description  Process $hEventContent which contains inventory-related events.
#               Incoming messages for the following events:
#               <ul>
#                   <li>InventoryBalanceChanged - includes stock level changes for a list of products
#                   <li>InventoryTrackingStarted - informs tracking was enabled for a product
#                   <li>InventoryTrackingStopped - informs tracking was disabled for a product
#               </ul>
#               Events will be parsed and the shop will be updated accordingly:
#               update stock levels or (un)set ZettleTrackingEnabled product attributes.
#----------------------------------------------------------------------------------------
# input        $hEventContent | contents of an incoming event message with the following keys:
#                                <ul>
#                                    <li>organizationUuid - string
#                                    <li>messageUuid - UUID v1 based on timestamp and messageId - string
#                                    <li>eventName - https://github.com/iZettle/api-documentation/blob/master/pusher-api/api-reference.md#supported-events - string
#                                    <li>messageId - unique UUID of the message - string
#                                    <li>payload - specific payload depending on the type of event.
#                                        see https://github.com/iZettle/api-documentation/blob/master/pusher-api/user-guides/subscriptions.md#payloads - hash
#                                    <li>timestamp - event timestamp in ISO-8601 format - string
#                                </ul>
#========================================================================================
sub incomingStockUpdate {
    my $self = shift;
    my ($hEventContent) = @_;
    my $MethodName = (caller(0))[3];

    my $Payload = $hEventContent->{'payload'};
    # Decode event payload (which is also encoded)
    $hEventContent->{'payload'} = decode_json($Payload) if defined $Payload;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();

    my $WebhookEventDTO = DE_EPAGES::Zettle::API::WebService::DTO::WebhookEventDTO->fromHash($hEventContent);

    #TODO: skip duplicate events comparing messageId field with stored values

    my $OrganizationUuid = $WebhookEventDTO->organizationUuid();
    my $EventName = $WebhookEventDTO->eventName();
    my $MerchantUuid = $Environment->getMerchantUUID();
    my $MessageId = $WebhookEventDTO->messageId();
    Error('InvalidOrgUUID', {
        'EventName' => $EventName,
        'OrganizationUuid' => $OrganizationUuid,
        'MessageId' => $MessageId
    }) unless $MerchantUuid eq $OrganizationUuid;

    my $hPayload = $WebhookEventDTO->payload;

    if($EventName eq 'InventoryBalanceChanged'){
        my $StockUpdateDTO = DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateDTO->fromHash($hPayload);

        # Skip incoming events triggered by updates sent to Zettle from the shop
        my $externalUuid = $StockUpdateDTO->externalUuid();
        return if defined $externalUuid && ($externalUuid eq $MerchantUuid);

        my $aBalanceBeforeDTO = $StockUpdateDTO->balanceBefore();
        my $aBalanceAfterDTO = $StockUpdateDTO->balanceAfter();

        # Retrieve Store location UUID
        my $LocationStoreUuid = $Environment->locationStoreUuid();

        # Loop balance changes and update stock levels, considering:
        # - Item count for both BalanceBefore and BalanceAfter shall be the same.
        # - Changes for a product in BalanceBefore array shall be found in the same position
        #   as changes in BalanceAfter array (BalanceBefore[0] relates to BalanceAfter[0]).
        for(my $idx=0; $idx < scalar @{$aBalanceBeforeDTO}; $idx++){
            my $VariationUuid;
            eval{
                my $BalanceBeforeDTO = @{$aBalanceBeforeDTO}[$idx];
                my $BalanceAfterDTO = @{$aBalanceAfterDTO}[$idx];

                # Ensure we only process updates for Store location
                my $BalanceBeforeLocation = $BalanceBeforeDTO->locationUuid();
                my $BalanceAfterLocation = $BalanceBeforeDTO->locationUuid();
                if ($BalanceBeforeLocation ne $LocationStoreUuid
                        || $BalanceAfterLocation ne $BalanceAfterLocation) {
                    GetLog->debug("$MethodName - ignoring balance update in unexpected location for event '$EventName' "
                                    ."(beforeLocationUUID: $BalanceBeforeLocation, afterLocationUUID: $BalanceAfterLocation)");
                    next;
                }

                # Ensure product/variation UUIDs match for BalanceBefore item and BalanceAfter item
                my $BalanceBeforeProductUuid = $BalanceBeforeDTO->productUuid();
                my $BalanceAfterProductUuid = $BalanceAfterDTO->productUuid();
                my $BalanceBeforeVariationUuid = $BalanceBeforeDTO->variationUuid();
                my $BalanceAfterVariationUuid = $BalanceAfterDTO->variationUuid();
                if ($BalanceBeforeProductUuid ne $BalanceAfterProductUuid
                        || $BalanceBeforeVariationUuid ne $BalanceAfterVariationUuid) {
                    GetLog->debug("$MethodName - ignoring balance update for unpaired UUIDs for event '$EventName' "
                                    ."- product (before=$BalanceBeforeProductUuid, after: $BalanceAfterLocation), "
                                    ."variation (before=$BalanceBeforeVariationUuid, after: $BalanceAfterVariationUuid)");
                    next;
                }

                $VariationUuid = $BalanceBeforeDTO->variationUuid();

                my $BalanceBefore = $BalanceBeforeDTO->balance();
                my $BalanceAfter = $BalanceAfterDTO->balance();

                my $BalanceDifference = $BalanceAfter-$BalanceBefore;

                my $Product = LoadProductByZettleUUID($Shop, $VariationUuid);

                my $CurrentStockLevel = $Product->get('StockLevel');
                # Only update if incoming stock level is different from current stock level (this prevents unnecessary updates)
                if($BalanceAfter != $CurrentStockLevel){
                    $Product->set({
                        'StockLevel' => $CurrentStockLevel + $BalanceDifference,
                    },
                    undef,
                    {
                        'IgnoreZettleUpdate' => 1,  # Add a additionalOptions used in hook OnBeforeUpdateProduct to prevent sending Zettle updates back to Zettle
                    });
                }
            };
            if(ExistsError()){
                my $Error = GetError();

                if($Error->code eq 'NoResults_ProductByZettleUUID'){
                    # This is not a real error, but a warning. Zettle could store products imported from a source different than the shop.
                    # If tracking were enabled for any of these products it would trigger this callback message with an unknown UUID for us.
                    # It could also indicate some malfunctioning in our integration so we log it. NOTE: Remove if not necessary anymore!
                    my $Logger = $self->_getLogger();
                    $Logger->logCommunication($MethodName, {
                        'Error' => 'NoResults_ProductByZettleUUID',
                        'ZettleUuid' => $VariationUuid,
                        'IncomingEvent' => $EventName,
                        'MessageId' => $MessageId
                    });
                } else {
                    $Error->addVars({
                        'IncomingEvent' => $EventName,
                        'MessageId' => $MessageId
                    });
                    LogError($Error);
                }
            }else{
                $Environment->lastStockSyncZettle(GetCurrentDateTime());
            }
        }
    }elsif($EventName eq 'InventoryTrackingStarted'){
        my ($TrackedProductId, $TrackedProductZettleUuid);
        eval{
            $TrackedProductZettleUuid = $hPayload->{'productUuid'};
            my $TrackedProduct = LoadProductByZettleUUID($Shop, $TrackedProductZettleUuid);
            $TrackedProductId = $TrackedProduct->id();

            my $TrackingEnabled = $TrackedProduct->get('ZettleTrackingEnabled');
            # We only react if tracking was disabled.
            # If tracking was already enabled stock levels were updated in real time so no need to export product stock level.
            # If tracking was disabled we need to ensure Zettle stock levels match shop stock levels.
            if ($TrackingEnabled == 0) {
                $TrackedProduct->set({'ZettleTrackingEnabled' => 1});
                # Adjust Zettle stock levels when product tracking is re-enabled
                my $aProducts = [ $TrackedProduct ];
                $self->_exportStockLevels($aProducts);
            }
        };
        if(ExistsError()){
            my $Error = GetError();
            if($Error->code ne 'NoResults_ProductByZettleUUID'){
                $Error->addVars({
                    'IncomingEvent' => $EventName,
                    'MessageId' => $MessageId,
                    'ZettleProductUuid' => $TrackedProductZettleUuid,
                    'ProductID' => $TrackedProductId,
                });
                LogError($Error);
            }
        }
    }elsif($EventName eq 'InventoryTrackingStopped'){
        my ($UntrackedProductId, $UntrackedProductZettleUuid);
        eval{
            $UntrackedProductZettleUuid = $hPayload->{'productUuid'};
            my $UntrackedProduct = LoadProductByZettleUUID($Shop, $UntrackedProductZettleUuid);
            if(defined $UntrackedProduct){
                $UntrackedProductId = $UntrackedProduct->id();
                $UntrackedProduct->set({'ZettleTrackingEnabled' => 0});
            }
        };
        if(ExistsError()){
            my $Error = GetError();
            # When a product is deleted from the shop inventory tracking is disabled and this generates an incoming request
            # This request refers the ZettleUUID of the deleted product which is not available anymore so we just ignore it
            if($Error->code ne 'NoResults_ProductByZettleUUID'){
                $Error->addVars({
                    'IncomingEvent' => $EventName,
                    'MessageId' => $MessageId,
                    'ZettleProductUuid' => $UntrackedProductZettleUuid,
                    'ProductID' => $UntrackedProductId,
                });
                LogError($Error);
            }
        }
    }else{
        GetLog->debug("$MethodName - ignoring unexpected event of type ".$EventName);
    }

    return;
}

#========================================================================================
# function     deleteProducts
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hDeleteResult = $self->deleteProducts($aProducts)
#----------------------------------------------------------------------------------------
# description  Delete given products from Zettle library.
#----------------------------------------------------------------------------------------
# syntax       $aProducts | list of products to delete | ref.array.DE_EPAGES::Product::API::Object::Product
#----------------------------------------------------------------------------------------
# return       $hDeleteResult | List of deleted products and list of failed deletions, with following keys:
#               <ul>
#                   <li>deleted - list of IDs of products successfully deleted from Zettle - ref.array.string
#                   <li>failed - list of IDs of products that couldn't be deleted - ref.array.string
#               </ul>| ref.hash
#========================================================================================
sub deleteProducts {
    my $self = shift;
    my ($aProducts) = @_;

    my $Environment = $self->getEnvironment();
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
    
    # Map of product data where key-value pair is: ZettleUUID => product ID
    my $hProductIdMap = {};
    # Compile ZettleUUIDs to delete (associating them with their product ID)
    foreach my $Product (@{$aProducts}) {
        my $ZettleUUID = $Product->get('ZettleUUID');
        # Skip products that don't have ZettleUUID
        next unless defined $ZettleUUID;
        $hProductIdMap->{$ZettleUUID} = $Product->id;
    }
    
    my @aZettleUUIDsToDelete = keys %{$hProductIdMap};
    return [] unless scalar @aZettleUUIDsToDelete;
    
    my $hDeleteResult = $ProductService->deleteProducts(\@aZettleUUIDsToDelete);
    
    # Convert ZettleUUIDs to product IDs using map previously compiled
    foreach my $idx ( qw ( deleted failed ) ) {
        my $aZettleUUIDs = $hDeleteResult->{$idx};
        $hDeleteResult->{$idx} = map { $hProductIdMap->{$_} } @{$aZettleUUIDs};
    }

    return $hDeleteResult;
}

#========================================================================================
# function     removeProductsZettleUUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->removeProductsZettleUUID($aProducts)
#----------------------------------------------------------------------------------------
# description  Removes the ZettleUUID attribute value from received products and their variants
#----------------------------------------------------------------------------------------
# syntax       $aProducts | List of products from which we want to delete the value of ZettleUUI | ref.array.DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub removeProductsZettleUUID {
    my $self = shift;
    my ($aProducts) = @_;

    foreach my $Product (@{$aProducts}) {
        $Product->set({'ZettleUUID' => undef});
        foreach my $SubProduct (@{$Product->get('SubProducts')}) {
            $SubProduct->set({'ZettleUUID' => undef});
        }
    }
    
}


#========================================================================================


#========================================================================================
# function     _ExportExceedsDailyLimit
# state        private
#----------------------------------------------------------------------------------------
# syntax       $IsExceeded = $self->_ExportExceedsDailyLimit($ShopID, $aProducts)
#----------------------------------------------------------------------------------------
# description  Check whether exporting the given items to Zettle would exceed API product export daily limit.
#----------------------------------------------------------------------------------------
# input        $ShopID | ID of the shop | int
# input        $aProducts | List of products to be exported to Zettle | ref.array.DE_EPAGES::Product::API::Object::Product
#----------------------------------------------------------------------------------------
# return       $IsExceeded | true if limit exceeded, false otherwise | int
#========================================================================================
sub _ExportExceedsDailyLimit {
    my ($ShopID, $aProducts) = @_;

    # Count the number of superproducts to export.
    # $aProducts may contain both superproducts and subproducts.
    # We create a map of superproduct paths, removing the 'SubProduct/XXX' from RelativePath.
    my %hProductPaths = map { (my $str  = $_->get('RelativePath')) =~ s/\/SubProducts\/.*//; $str => 1 } @{$aProducts};
    my $ExportItemCount = scalar keys %hProductPaths;
    return 1 if ($ExportItemCount > API_MAX_PRODUCTS_PER_DAY);

    #TODO: sum of items (itemcount) of registered exports in zettleproductexport for the last 24 hours

    return 0;
}

#========================================================================================
# function     _exportLastProductDeletions
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hDeleteResult = $self->_exportLastProductDeletions($LastProductExport)
#----------------------------------------------------------------------------------------
# description  Check products deleted from the shop since $LastProductExport and send
#               deletion of such products to Zettle.
#----------------------------------------------------------------------------------------
# input        $LastProductExport | a date and time reference to retrieve deletion of products | Datetime
#----------------------------------------------------------------------------------------
# return       $hDeleteResult | List of deleted products and list of failed deletions, with following keys:
#               <ul>
#                   <li>deleted - list of ZettleUUIDs of products successfully deleted from Zettle - ref.array.string
#                   <li>failed - list of ZettleUUIDs of products that couldn't be deleted - ref.array.string
#               </ul>| ref.hash
#========================================================================================
sub _exportLastProductDeletions {
    my $self = shift;
    my ($LastProductExport) = @_;

    my $Environment = $self->getEnvironment();
    
    # Retrieve products deleted since last export
    my $ZettleDeleteIterator = $Environment->getDeletedObjects($LastProductExport);
    # $ahZettleDeletes contains a list of ZettleDelete hashes which hold information of objects
    # deleted from the shop and their ZettleUUIDs (required to delete them from Zettle).
    # A ZettleDelete hash contains the following keys:
    # <ul>
    #     <li>GUID - object GUID - string
    #     <li>SiteID - shop ID - int
    #     <li>ClassID - class ID - int
    #     <li>PathFromSite - path of object relative to site - string
    #     <li>ZettleUUID - Zettle UUID of the object - string
    #     <li>DeletedOn - date of deletion - datetime
    # </ul>
    my @ahZettleDeletes = ();
    while(my $hZettleDelete = <$ZettleDeleteIterator>){
        push(@ahZettleDeletes, $hZettleDelete);
    }

    my $hDeleteResult;
    #TODO: use different 'eval' blocks for each API operation
    eval {
        my $ShopPath = $Environment->getSite()->get('Path');
        my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);

        # [STEP] Delete superproducts (with or without variations)
        my %deletedParents = map { $_->{'PathFromSite'} => $_ } grep { $_->{'PathFromSite'} !~ /\/SubProducts\// } @ahZettleDeletes;
        if(scalar values %deletedParents){
            my @aDeleteZettleUUIDs = map { $_->{'ZettleUUID'} } values %deletedParents;
            $hDeleteResult = $ProductService->deleteProducts(\@aDeleteZettleUUIDs);
        }

        # [STEP] Compile deleted variations whose superproducts were not deleted too
        my %deletedVariations = ();
        foreach my $hZettleDelete ( @ahZettleDeletes ){
            my $Path = $hZettleDelete->{'PathFromSite'};
            # Skip superproducts
            next if $Path !~ /SubProducts/;
            
            $Path =~ s/\/SubProducts\/.*//;
            # Skip subproducts if its superproduct is included in %deletedParents (subproducts will be cascade deleted in Zettle)
            next if exists $deletedParents{$Path};

            # Create a new array hash entry indexed by superproduct path
            $deletedVariations{$Path} //= [];
            push(@{$deletedVariations{$Path}}, $hZettleDelete);
        }

        # If all variations were deleted for a product we have to create a fake variation
        my $aProductsForFakeVariations = ();
        # If only some variations of a superproduct were deleted, we have to "delete" them using an HTTP PUT request
        my $aProductsForPutUpdate = ();
        foreach my $ParentPath (keys %deletedVariations){
            my $ProductFullPath = "$ShopPath/$ParentPath";
            next unless(ExistsObjectByPath($ProductFullPath));

            my $Product = LoadObjectByPath($ProductFullPath);
            my $SubProductCount = $Product->countSubProducts();
            if($SubProductCount > 0){
                push(@{$aProductsForPutUpdate}, $Product);
            }else{
                push(@{$aProductsForFakeVariations}, $Product);
            }
        }

        my $ZettleUUID;
        my $HttpResponse;
        my $ETag;

        # [STEP] Use updateProduct (HTTP PUT) to delete some, but not all, variations of a product
        foreach my $ProductForUpdate (@{$aProductsForPutUpdate}){
            $ZettleUUID = $ProductForUpdate->get('ZettleUUID');
            $HttpResponse = $ProductService->getProduct($ZettleUUID);
            $ETag = $HttpResponse->header('ETag');
            $ProductService->updateProduct($ProductForUpdate, $ETag);
        }

        # [STEP] Use updateProduct (HTTP PUT) to create a fake variation for a product without subproducts
        foreach my $ProductForFake (@{$aProductsForFakeVariations}){
            $ZettleUUID = $ProductForFake->get('ZettleUUID');
            $HttpResponse = $ProductService->getProduct($ZettleUUID);
            $ETag = $HttpResponse->header('ETag');
            $ProductService->updateProduct($ProductForFake, $ETag);
        }
    };
    if(ExistsError()){
        LogError(GetError());
        #TODO: augment $hDeleteResult with error details
    }

    #TODO: complete $hDeleteResult with result from calls to $ProductService->updateProduct() above
    
    $hDeleteResult //= {
        'deleted' => [],
        'failed' => []
    };
    return $hDeleteResult;
}

#========================================================================================
# function     _exportImages
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hImageUploadResults = $self->_exportImages($aProducts, $ImageFormat)
#----------------------------------------------------------------------------------------
# description  Export to Zettle images for given products in the specified format.
#----------------------------------------------------------------------------------------
# input        $aProducts | list of products objects | ref.array.DE_EPAGES::Product::API::Object::Product
# input        $ImageFormat | optional. name of the image format to be processed,
#                              any of: ImageLarge, ImageMedium, ImageSmall, ImageThumbnail
#                              If argument not present, all product images will be exported | string
#----------------------------------------------------------------------------------------
# return       $hImageUploadResults | a map of ImageUploadResult objects indexed by product GUID.
#                                      See <DE_EPAGES::Zettle::API::Object::ImageUploadResult> | ref.hash.array.object
#========================================================================================
sub _exportImages {
    my $self = shift;
    my ($aProducts, $ImageFormat) = @_;
    
    my $Environment = $self->getEnvironment();
    my $ImageService = DE_EPAGES::Zettle::API::WebService::ImageService->new($Environment);
    my $hImageUrls = GetGuidImageMap($aProducts, $ImageFormat);

    # Merge the list of image URLs of each product
    my @aImageUrls;
    foreach my $aImageUrlSubList (values %{$hImageUrls}) {
        push(@aImageUrls, @{$aImageUrlSubList});
    }

    return {} if scalar(@aImageUrls) == 0;

    my $hImageUploadResults;

    my $ImagesCount = scalar @aImageUrls;
    for (my $StartIndex=0; $StartIndex<$ImagesCount; $StartIndex+=MAX_IMAGES_PER_EXPORT) {
        sleep(IMAGES_EXPORT_WAIT_INTERVAL) unless ($StartIndex == 0);
        
        # Batch Products
        my $EndIndex = ($StartIndex + MAX_IMAGES_PER_EXPORT >= $ImagesCount) ? $ImagesCount-1 : $StartIndex + MAX_IMAGES_PER_EXPORT -1;
        my @aBatchImageUrls = @aImageUrls[$StartIndex..$EndIndex];

        # Upload Images
        my $BulkImageUploadResponseDTO = $ImageService->bulkUploadImages(\@aBatchImageUrls);
        my $hBatchImageUploadResults = GetImageUploadResults($BulkImageUploadResponseDTO, $hImageUrls);

        # Merge batch results
        if (defined($hImageUploadResults)) {
            $hImageUploadResults = { %$hImageUploadResults, %$hBatchImageUploadResults };
        } else {
            $hImageUploadResults = $hBatchImageUploadResults;
        }
    }

    return $hImageUploadResults;
}

#========================================================================================
# function     _updateShopProductsUUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_updateShopProductsUUID($aProductResponseDTO)
#----------------------------------------------------------------------------------------
# input        $aProductResponseDTO | Array of ProductResponseDTO (see DE_EPAGES::Zettle::API::WebService::DTO::ProductResponseDTO) | ref.array.object
#----------------------------------------------------------------------------------------
# description  Use product info retrievd from Zettle to fill in product ZettleUUID attribute for products previously exported to Zettle.
#               The goal is to avoid generating product duplicates in Zettle given that UUID information is the unique identifier used in Zettle API requests.
#               This let's us reuse previously generated UUIDs instead of generating new ones, which would end up with duplicates in Zettle.
#----------------------------------------------------------------------------------------
# return       $hUpdatedProducts | hash of updated products where key is product/variation ID and value is updated UUID | ref.hash
#========================================================================================
sub _updateShopProductsUUID {
    my $self = shift;
    my ($aProductResponseDTO) = @_;

    # Hash of product UUID updates
    my $hUpdatedProducts = {};
    foreach my $ProductResponseDTO (@{$aProductResponseDTO}) {
        my $ZettleProductUuid = $ProductResponseDTO->uuid();
        # If product was exported from the shop, externalReference will contain product GUID
        my $ZettleExternalReference = $ProductResponseDTO->externalReference();
        if(defined $ZettleExternalReference && ExistsObjectByGUID($ZettleExternalReference)){
            eval {
                my $Product = LoadObjectByGUID($ZettleExternalReference);
                TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');
                $Product->set({'ZettleUUID' => $ZettleProductUuid});
                $hUpdatedProducts->{$Product->id} = $ZettleProductUuid;
                
                # Also update UUID for product variations
                my $ahRawVariations = $ProductResponseDTO->variations();
                foreach my $hRawVariation (@{$ahRawVariations}) {
                    my $ZettleVariationUuid = $hRawVariation->{'uuid'};
                    my $ZettleVariationOptions = $hRawVariation->{'options'};
                    
                    # NOTE: Zettle variations don't have externalReference field so the best we can do is match by variation attribute options.
                    my $SubProductMatch = $self->_matchSubProductByOptions($Product, $ZettleVariationOptions);
                    if (defined $SubProductMatch) {
                        $SubProductMatch->set({'ZettleUUID' => $ZettleVariationUuid});
                        $hUpdatedProducts->{$SubProductMatch->id} = $ZettleVariationUuid;
                    }
                }
            };
            if (ExistsError()) {
                my $Error = GetError();
                LogError($Error);
            }
        }
    }

    return $hUpdatedProducts;
}

#========================================================================================
# function     _matchSubProductByOptions
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SubProduct = $self->_matchSubProductByOptions($Product, $ZettleVariationOptions)
#----------------------------------------------------------------------------------------
# input        $Product | a super product | DE_EPAGES::Product::API::Object::Product
# input        $ZettleVariationOptions | attribute values for a zettle variation | ref.array.hash
#               [
#                   { 'name' => 'Colour', 'value' => 'Blue' },
#                   { 'name' => 'Size', 'value' => 'M' },
#                   { 'name' => 'Material', 'value' => 'Cotton' },
#               ]
#----------------------------------------------------------------------------------------
# description  Search SubProducts of given $Product and return the one whose variation attribute options match $ZettleVariationOptions.
#----------------------------------------------------------------------------------------
# return       $SubProduct | a SubProduct that matches given options or undef | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub _matchSubProductByOptions {
    my $self = shift;
    my ($Product, $ZettleVariationOptions) = @_;

    foreach my $SubProduct (@{$Product->get('SubProducts')}) {
        my $SubProductOptions = BuildVariantOptions($SubProduct);
        # We need to specify full Compare package to avoid perl critic complaining about PerlTools not 'use'd
        return $SubProduct if Data::Compare::Compare($SubProductOptions, $ZettleVariationOptions);
    }

    return;
}

#========================================================================================
# function     _disableInventoryTracking
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_disableInventoryTracking($aProducts)
#----------------------------------------------------------------------------------------
# description  Disable InventoryTracking for given products.
#----------------------------------------------------------------------------------------
# input        $aProducts | List of products | ref.array.DE_EPAGES::Product::API::Object::Product
#----------------------------------------------------------------------------------------
# return       $Updated | TRUE if product tracking was successfully updated | boolean
#========================================================================================
sub _disableInventoryTracking {
    my $self = shift;
    my ($aProducts) = @_;

    my $TrackingStatus = 'disable';
    return $self->_updateInventoryTracking($aProducts, $TrackingStatus);
}

#========================================================================================
# function     _enableInventoryTracking
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_enableInventoryTracking($aProducts)
#----------------------------------------------------------------------------------------
# description  Enable InventoryTracking for given products.
#----------------------------------------------------------------------------------------
# input        $aProducts | List of products | ref.array.DE_EPAGES::Product::API::Object::Product
#----------------------------------------------------------------------------------------
# return       $Updated | TRUE if product tracking was successfully updated | boolean
#========================================================================================
sub _enableInventoryTracking {
    my $self = shift;
    my ($aProducts) = @_;

    my $TrackingStatus = 'enable';
    return $self->_updateInventoryTracking($aProducts, $TrackingStatus);
}

#========================================================================================
# function     _updateInventoryTracking
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_updateInventoryTracking($aProducts, $TrackingStatus)
#----------------------------------------------------------------------------------------
# description  Update InventoryTracking for given products. Either enable or disable tracking in Zettle.
#               Also update ZettleTrackingEnabled attribute for each processed product.
#----------------------------------------------------------------------------------------
# input        $aProducts | List of products | ref.array.DE_EPAGES::Product::API::Object::Product
# input        $TrackingStatus | Any of: enable, disable | string
#----------------------------------------------------------------------------------------
# return       $Updated | TRUE if product tracking was successfully updated | boolean
#========================================================================================
sub _updateInventoryTracking {
    my $self = shift;
    my ($aProducts, $TrackingStatus) = @_;

    return 0 unless ($TrackingStatus eq 'enable') || ($TrackingStatus eq 'disable');
    
    # Ensure tracking operations are not executed on variations (Zettle tracking is handled at parent level)
    my @aSuperProducts = grep { not defined $_->get('SuperProduct') } @{$aProducts};
    my @aProductUuid = map { $_->get('ZettleUUID') } grep { defined $_->get('ZettleUUID') } @aSuperProducts;

    return 1 unless scalar @aProductUuid;

    my $InventoryService = $self->_getInventoryService();
    my $Updated = $InventoryService->bulkUpdateProductTracking(\@aProductUuid, $TrackingStatus);

    if ($Updated) {
        # If $TrackingStatus eq 'enable' we subscribe to InventoryTrackingStarted event and it will generate an incoming callback for each of the products tracked.
        # Even though, we force here TrackingEnabled in all products given that asynchronous messages might take a while to be received
        # This would cause stock levels not to be exported for some products
        my $TrackingEnabled = ($TrackingStatus eq 'enable') ? 1 : 0;
        $_->set({'ZettleTrackingEnabled' => $TrackingEnabled}) foreach @{$aProducts};
    }

    return $Updated;
}

#========================================================================================
# function     updateAllLocationUUID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->updateAllLocationUUID($Environment)
#----------------------------------------------------------------------------------------
# description  Retrieve all location UUIDs from Zettle and store them in ZettleEnvironment.
#----------------------------------------------------------------------------------------
# input        $Environment | a Zettle environment object | DE_EPAGES::Zettle::API::Object:ZettleEnvironment
#----------------------------------------------------------------------------------------
# return       $aLocationDTO | List of locations.
#                               See DE_EPAGES::Zettle::API::WebService::DTO::LocationDTO | ref.array.object
#========================================================================================
sub updateAllLocationUUID {
    my $self = shift;
    
    my $Environment = $self->getEnvironment();

    my $InventoryService = $self->_getInventoryService();
    my $aLocationDTO = $InventoryService->getAllLocations();
    foreach my $LocationDTO (@{$aLocationDTO}){
        my $LocationUuid = $LocationDTO->inventoryUuid();
        if($LocationDTO->name() eq 'BIN'){
            $Environment->locationBinUuid($LocationUuid);
        }elsif($LocationDTO->name() eq 'SOLD'){
            $Environment->locationSoldUuid($LocationUuid);
        }elsif($LocationDTO->name() eq 'STORE'){
            $Environment->locationStoreUuid($LocationUuid);
        }elsif($LocationDTO->name() eq 'SUPPLIER'){
            $Environment->locationSupplierUuid($LocationUuid);
        }
    }

    return $aLocationDTO;
}

#========================================================================================
# function     _exportStockLevels
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hExportStockBulk = $self->_exportStockLevels($aProducts)
#----------------------------------------------------------------------------------------
# description  Export stock levels for given products.
#               Current balance reference will be retrieved from Zettle.
#----------------------------------------------------------------------------------------
# input        $aProducts | List of products | ref.array.object
#----------------------------------------------------------------------------------------
# return       $hExportStockBulk | List of updated variations and list of failed updates, with following keys:
#               <ul>
#                   <li>updated - list of ZettleUUIDs of variations successfully updated in Zettle - ref.array.string
#                   <li>failed - list of ZettleUUIDs of variations that failed to be updated - ref.array.string
#               </ul>| ref.hash
#========================================================================================
sub _exportStockLevels {
    my $self = shift;
    my ($aProducts) = @_;

    my $hBalanceUpdate = {};
    
    # Retrieve current balances from Zettle to calculate differences with shop stock levels
    my $aProductInventoryDTO = $self->_getZettleBalances($aProducts);

    foreach my $Product (@{$aProducts}) {
        my $SuperProduct = $Product->get('SuperProduct') // $Product;

        # Only export stock levels for tracked products (tracking is handled at superproduct level)
        next unless $SuperProduct->get('ZettleTrackingEnabled') == 1;

        my $SuperProductUuid = $SuperProduct->get('ZettleUUID');
        next unless defined $SuperProductUuid;

        my $aSubProducts = $Product->get('SubProducts');
        if (scalar @{$aSubProducts}) {
            # Calculate stock update for each subproduct: difference between shop stock and zettle stock
            foreach my $SubProduct (@{$aSubProducts}) {
                my $VariantUuid = $SubProduct->get('ZettleUUID');
                #TODO: should we raise/log an error if no ZettleUUID
                # Variant may not have ZettleUUID if it was not exported to Zettle for some reason
                next unless defined $VariantUuid;
                my $StockDelta = $self->_getProductStockDelta($SubProduct, $aProductInventoryDTO);
                $hBalanceUpdate->{$SuperProductUuid}->{$VariantUuid} = $StockDelta;
            }
        } else {
            # If no SubProducts, we're dealing with either a master product or with a variation
            # Either way, we'll use product's ZettleUUID as VariantUUID
            my $VariantUuid = $Product->get('ZettleUUID');
            #TODO: should we raise/log an error if no ZettleUUID
            # Product may not have ZettleUUID if it was not exported to Zettle for some reason
            next unless defined $VariantUuid;
            my $StockDelta = $self->_getProductStockDelta($Product, $aProductInventoryDTO);
            $hBalanceUpdate->{$SuperProductUuid}->{$VariantUuid} = $StockDelta;
        }
    }

    my $InventoryService = $self->_getInventoryService();
    my $hExportStockBulk = $InventoryService->bulkUpdateProductBalance($hBalanceUpdate);
    
    my $Environment = $self->getEnvironment();
    $Environment->lastStockSync(GetCurrentDateTime());

    return $hExportStockBulk;
}

#========================================================================================
# function     _getZettleBalances
# state        private
#----------------------------------------------------------------------------------------
# syntax       $aProductInventoryDTO = $self->_getZettleBalances($aProducts, $LocationUuid);
#----------------------------------------------------------------------------------------
# description  Retrieve variation balances for given products from given Zettle location.
#----------------------------------------------------------------------------------------
# input        $aProducts | List of products | ref.array.object
# input        $LocationUuid | optional. Location UUID (default: UUID for Store location) | string
#----------------------------------------------------------------------------------------
# return       $aProductInventoryDTO | List of variation balances | ref.array.DE_EPAGES::Zettle::API::WebService::DTO::ProductInventoryDTO
#========================================================================================
sub _getZettleBalances {
    my $self = shift;
    my ($aProducts) = @_;

    my $InventoryService = $self->_getInventoryService();

    # Compile a unique list of ZettleUUIDs of parent products
    # $aProducts may contain: master products, superproducts and/or variations
    my $hProductZettleUUIDs = {};
    foreach my $Product (@{$aProducts}) {
        # Since Zettle inventory is retrieved for parent UUID,
        # if we're dealing with a variation we need to use its parent ZettleUUID
        my $SuperProduct = $Product->get('SuperProduct') // $Product;

        # Skip products which don't have tracking enabled
        next unless $SuperProduct->get('ZettleTrackingEnabled') == 1;

        my $ZettleUuid = $SuperProduct->get('ZettleUUID');
        next unless defined $ZettleUuid;

        $hProductZettleUUIDs->{$ZettleUuid} = 1;
    }

    # Request product stock levels from Zettle
    my @ProductZettleUUIDs = keys %{$hProductZettleUUIDs};
    my $aProductInventoryDTO = $InventoryService->getProductListInventory(\@ProductZettleUUIDs);

    return $aProductInventoryDTO;
}

#========================================================================================
# function     _getProductStockDelta
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_getProductStockDelta($Product, $aProductInventoryDTO)
#----------------------------------------------------------------------------------------
# description  Calculate difference between shop stock level and zettle stock level.
#               Search $Product ZettleUUID in $aProductInventoryDTO in order to calculate difference.
#               If ZettleUUID is not present in hash return $Product StockLevel.
#               If it's present, return ShopStock - ZettleStock.
#----------------------------------------------------------------------------------------
# input        $Product | a Product object | DE_EPAGES::Product::API::Object::Product
# input        $aProductInventoryDTO | List of variation balances | ref.array.DE_EPAGES::Zettle::API::WebService::DTO::ProductInventoryDTO
#----------------------------------------------------------------------------------------
# return       $StockDelta | Difference between Zettle stock and shop stock | integer
#========================================================================================
sub _getProductStockDelta {
    my $self = shift;
    my ($Product, $aProductInventoryDTO) = @_;

    my $ZettleUuid = $Product->get('ZettleUUID');
    my $ShopStock = $Product->get('StockLevel') // 0;
    
    # Search $aProductInventoryDTO for a variant that matches $ZettleUuid
    my $ZettleStock = 0;
    foreach my $ProductInventoryDTO (@{$aProductInventoryDTO}){
        if($ProductInventoryDTO->variantUuid() eq $ZettleUuid){
            $ZettleStock = $ProductInventoryDTO->balance();
            last;
        }
    }

    my $StockDelta = $ShopStock - $ZettleStock;

    return $StockDelta;
}

#========================================================================================
# function     _removeEnvironmentSubscription
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_removeEnvironmentSubscription($Environment)
#----------------------------------------------------------------------------------------
# description  Delete subscription from Zettle and remove subscription data from environment.
#----------------------------------------------------------------------------------------
# input        $Environment | Zettle Environment | Object
#========================================================================================
sub _removeEnvironmentSubscription {
    my $self = shift;

    my ($Environment) = @_;

    # Delete subscriptions from Zettle
    my $InventorySubscriptionUUID = $Environment->getInventorySubscriptionUUID();
    if(defined($InventorySubscriptionUUID)) {
        my $PusherService = DE_EPAGES::Zettle::API::WebService::PusherService->new($Environment);
        eval {
            $PusherService->deleteSubscription($InventorySubscriptionUUID);
        };
        if(ExistsError()){ LogError(GetError()); }
    }
    
    # Remove Subscription data from environment
    $Environment->removeSubscriptionData();
}

#========================================================================================
sub _getTokenHelper {
    my $self = shift;
    my ($Environment) = $self->getEnvironment();

    $self->{'TokenHelper'} //= $self->_createTokenHelper($Environment);
    return $self->{'TokenHelper'};
}

#========================================================================================
sub _createTokenHelper {
    my $self = shift;
    my ($Environment) = @_;
    $self->{'TokenHelper'} //= DE_EPAGES::Zettle::API::TokenHelper->new($Environment);
    return $self->{'TokenHelper'};
}

#========================================================================================
sub _getStatusFromState {
    my $self = shift;
    my ($State) = @_;

    my $ExportStatus;
    if($State eq 'IMPORTING'){
        $ExportStatus = EXPORT_STATUS->{RUNNING};
    }elsif($State eq 'FINISHED_SUCCESS'){
        $ExportStatus = EXPORT_STATUS->{COMPLETED};
    }elsif($State eq 'FINISHED_FAILED'){
        $ExportStatus = EXPORT_STATUS->{FAILED};
    }

    return $ExportStatus;
}

#========================================================================================
sub _getLogger {
    my $self = shift;

    $self->{'Logger'} //= DE_EPAGES::Zettle::API::Logger->new();
    return $self->{'Logger'};
}

#========================================================================================
sub _getInventoryService {
    my $self = shift;

    my $Environment = $self->getEnvironment();
    $self->{'InventoryService'} //= DE_EPAGES::Zettle::API::WebService::InventoryService->new($Environment);
    return $self->{'InventoryService'};
}


1;
